import traceback
import inspect
import logging
from logging import getLogger
from time import time, localtime, strftime
import asyncio
import os
import sys
from socket import gethostname, getfqdn

from urllib.parse import unquote, parse_qs
from requests.structures import CaseInsensitiveDict


from jsonrpc_client import ProxyException
from jsonrpc_client import BaseResponse as JsonRpcClientBaseResponse
import json2
from checkerslib import Checker, DictChkr, StrChkr, IntChkr, AnyChkr, EnumChkr
from checkerslib import ListChkr  # noqa
from checkerslib import UNDEFINED

from .exceptions import BaseRPCException, MethodNotFound, MissingArgument
from .exceptions import SessionRequired, MethodHelpNotFound
from .Server import Server
from .docs_generator import DocsGenerator

from .wrapper_classes import generatorTypes, ResponseWrap

# následující import je potřeba ponechat - spoléhají se na něj některé projekty
from .wrapper_classes import make_response_obj  # noqa F401

try:  # Python 3.5+
    from http import HTTPStatus
except ImportError:
    from http import client as HTTPStatus


log = getLogger("jsonrpc")
time_log = getLogger("jsonrpc.time")
exc_log = getLogger("jsonrpc.exc")

try:
    # pylint: disable=import-error
    import bson
except ImportError:
    bson = None
    log.warning("Bson not available")
# endtry

try:
    # pylint: disable=import-error
    import uwsgi
except ImportError:
    uwsgi = None
    log.warning("uwsgi not available")
# endtry


dicts = (dict, CaseInsensitiveDict)


class RPC:
    """
    Aplikační logika, která se napojí na WSGI rozhrani.
    """

    app_name = None
    app_version = "0.0.0"
    app_base_url = None
    duality_master = None

    response_headers = [
        ("Content-Type", "application/json"),
        # ("Author", "lukoko@lukoko.name"),
        ("Access-Control-Allow-Origin", "*"),
        ("Access-Control-Allow-Methods", "POST, GET, OPTIONS"),
        (
            "Access-Control-Allow-Headers",
            "Reply-OK-Only, Content-Type, Origin, X-Requested-With, Accept",
        ),
        ("Connection", "Close"),
    ]
    callback_params = ("_callback", "callback", "_")

    # pylint: disable=too-many-arguments,too-many-locals,too-many-statements
    def __init__(
        self,
        request_headers=False,
        request_session=False,
        niceOutput=False,  # pylint: disable=invalid-name
        enable_introspective=True,
        log_time=True,
        time_critical=10,
        time_warning=1,
        config=None,
        bson_support=False,
        environment="dev",
        silent_methods=None,
        nullstream_append_terminator=True,
        add_content_length=False,
        enable_help=True,
        help_lang="cs",
        enable_help_call=True,
    ):
        self._servers = []
        self._methods = {
            "system_list_methods": self._list_methods,
            "system_method_help": self._method_help,
            "system_method_info": self._method_info,
            "system_method_sig": self._method_sig,
            "system_method_sig2": self._method_sig2,
            "system_method_json_schema": self._method_json_schema,
            "system_method_response_json_schema": (
                self._method_response_json_schema
            ),
            "system_status": self._method_status,
            "system_alive": self._method_alive,
            "system_multicall": self._multicall,
            "system_methods_info": self._methods_info,
            "openapi_v3": self._openapi_v3,
        }
        self._internal_methods = list(self._methods.keys())
        self.config = config

        self.enable_help = enable_help
        self.docs_generator = None
        if self.enable_help:
            self._methods[""] = self._method_index
            self.docs_generator = DocsGenerator(
                app=self,
                config=self.config,
                enable_help_call=enable_help_call,
                help_lang=help_lang,
            )
        self._default_method = None
        self.session_callback = False
        self.auth_callback = False
        self.session_blacklist = ["login"]
        self.request_headers = request_headers
        self.request_session = request_session
        self.niceOutput = niceOutput  # pylint: disable=invalid-name
        self.enable_introspective = enable_introspective
        self.log_time = log_time
        self.time_critical = time_critical
        self.time_warning = time_warning
        self.bson = bson is not None and bson_support is True
        self.environment = environment
        self.silent_methods = silent_methods or [
            "OPTIONS",
        ]
        self.nullstream_append_terminator = nullstream_append_terminator
        self.add_content_length = add_content_length
        if self.bson:
            log.critical(
                "Enabling bson support is experimental and can leed to "
                "unexpected behavior"
            )

        if self.config and self.duality_master is None:
            self.duality_master = self.config.getboolean(
                "uwsgi", "duality_master", False
            )
        # endif

        if self.config and self.config.has_section("rpc"):
            self.app_name = self.config.get(
                "rpc", "doc_app_name", self.app_name
            )
            self.request_session = self.config.getboolean(
                "rpc", "request_session", request_session
            )
            self.request_headers = self.config.getboolean(
                "rpc", "request_headers", request_headers
            )
            self.niceOutput = self.config.getboolean(
                "rpc", "nice_output", niceOutput
            )
            self.enable_introspective = self.config.getboolean(
                "rpc", "introspective", enable_introspective
            )
            self.time_warning = self.config.getint(
                "rpc", "time_warning", time_warning
            )
            self.time_critical = self.config.getint(
                "rpc", "time_critical", time_critical
            )
            self.environment = self.config.get(
                "rpc", "environment", environment
            )
            self.silent_methods = self.config.getlist(
                "rpc", "silent_methods", self.silent_methods
            )
            self.nullstream_append_terminator = self.config.getboolean(
                "rpc",
                "nullstream_append_terminator",
                self.nullstream_append_terminator,
            )
            self.add_content_length = self.config.getboolean(
                "rpc", "add_content_length", add_content_length
            )
            self.app_base_url = (
                self.config.get("rpc", "doc_base_url", None) or None
            )

        # endif

        if self.app_base_url is None:
            port = ":80"
            if self.config:
                port = self.config.get("uwsgi", "http-socket", None) or port
            self.app_base_url = f"http://{getfqdn()}{port}"

        if self.environment == "prod":
            self.niceOutput = False

        version = os.getcwd() + "/doc/version"

        if os.path.exists(version):
            with open(version, encoding="utf8") as f:
                self.app_version = f.read().strip()

    # enddef

    # pylint: disable=too-many-return-statements,too-many-branches
    # pylint: disable=too-many-statements
    def __call__(self, environment, start_response):
        # tady musí být .get protože v multicallu REQUEST_METHOD není
        start_response = ResponseWrap(start_response)

        if environment.get("REQUEST_METHOD") in self.silent_methods:
            response_headers = self.response_headers[:]
            if self.add_content_length:
                response_headers.append(("Content-Length", "0"))
            response_headers.append(("X-Target-Hostname", gethostname()))
            start_response(self._response_code(200), response_headers)
            yield b""
            return
        content_type = "application/json"
        jsonp = False
        reply_ok_only = False
        handler_name = None
        try:
            handler_name = environment["PATH_INFO"][1:]
            is_system_method = (
                handler_name.startswith("system_") or handler_name == ""
            )

            if is_system_method and not self.enable_introspective:
                # timhle vypinam internal metody pokud je tak receno v cfg
                raise MethodNotFound(handler_name)
            # endif

            handler = self._dispatch(handler_name)

            silent_methods = getattr(handler, "silent_methods", None)
            if (
                silent_methods
                and environment.get("REQUEST_METHOD") in silent_methods
            ):
                response_headers = self.response_headers[:]
                if self.add_content_length:
                    response_headers.append(("Content-Length", "0"))
                response_headers.append(("X-Target-Hostname", gethostname()))
                start_response(self._response_code(200), response_headers)
                yield b""
                return
            # endif

            get_raw_post_data = getattr(handler, "request_raw_post_data", None)

            handler_params, content_type, raw_post_data = self._query_params(
                environment, get_raw_post_data
            )
            log.debug("Called:%s?%s", handler_name, handler_params)

            reply_ok_only = bool(environment.get("HTTP_REPLY_OK_ONLY", False))

            for callback_param in self.callback_params:
                if callback_param in handler_params:
                    if not jsonp:
                        jsonp = handler_params[callback_param]
                    del handler_params[callback_param]
            # in py3 dict is ordered so this can be done
            response_headers = CaseInsensitiveDict(self.response_headers)
            if jsonp:
                response_headers["Content-Type"] = "text/javascript"
            # endif

            if not is_system_method and (
                self.request_headers
                or getattr(handler, "request_headers", None)
            ):
                headers = {
                    k[5:].lower().replace("_", "-"): v
                    for k, v in environment.items()
                    if k.startswith("HTTP_")
                }
                headers["remote-addr"] = environment.get("REMOTE_ADDR")
                handler_params["headers"] = headers
            # endif

            if get_raw_post_data:
                handler_params["raw_post_data"] = raw_post_data
            # endif

            request_raw_environment = getattr(
                handler, "request_raw_environment", None
            )
            if request_raw_environment is True:
                handler_params["raw_environment"] = {
                    k: environment[k]
                    for k in environment.keys()
                    if isinstance(environment[k], (str, bytes, int, bool))
                    # if (
                    #     not k.startswith("wsgi.") and
                    #     not k.startswith("wrapper")
                    # )
                }
            # endif

            response = {
                "status": 200,
                "status_message": "OK",
            }
            if not is_system_method:
                if (
                    self.session_callback
                    and handler_name not in self.session_blacklist
                ):
                    if "session" not in handler_params:
                        raise SessionRequired()
                    # pylint: disable=not-callable
                    user_info = self.session_callback(
                        session=handler_params["session"]
                    )
                    if log.isEnabledFor(logging.INFO):
                        log.info(
                            "%s : %s (%s)",
                            user_info,
                            handler_name,
                            json2.dumps(handler_params, nice=self.niceOutput),
                        )
                    if self.request_session:
                        handler_params["session"] = user_info
                    else:
                        del handler_params["session"]
                elif self.auth_callback:
                    handler_params["_method"] = handler_name
                    # pylint: disable=not-callable
                    handler_params = self.auth_callback(handler_params)
                    del handler_params["_method"]
                else:
                    if log.isEnabledFor(logging.INFO):
                        log.info(
                            "%s (%s)",
                            handler_name,
                            json2.dumps(handler_params, nice=self.niceOutput),
                        )
            if self.log_time:
                start_t = time()
            # endif
            resp = handler(**handler_params)
            is_generator = isinstance(resp, tuple(generatorTypes))
            log.debug("response is generator:%r", is_generator)

            if self.log_time:
                req_t = time() - start_t
                if req_t > self.time_critical:
                    func = time_log.fatal
                elif req_t > self.time_warning:
                    func = time_log.warning
                else:
                    func = time_log.debug

                func(
                    "Response takes %0.2fs for %s?%s",
                    req_t,
                    handler_name,
                    handler_params,
                )
            # endif

            if hasattr(resp, "headers") and resp.headers:
                if (
                    isinstance(resp, JsonRpcClientBaseResponse)
                    or self.add_content_length is False
                ):
                    response_headers.update(
                        {
                            k: v
                            for k, v in (
                                resp.headers.items()
                                if isinstance(resp.headers, dicts)
                                else resp.headers
                            )
                            if k.lower() != "content-length"
                        }
                    )
                elif isinstance(resp.headers, dicts):
                    response_headers.update(resp.headers)
                else:
                    response_headers.update(dict(resp.headers))
            elif hasattr(resp, "response_headers"):
                if isinstance(resp.response_headers, dicts):
                    response_headers.update(resp.response_headers)
                else:
                    response_headers.update(dict(resp.response_headers))

            if hasattr(handler, "raw_response") and handler.raw_response:
                if hasattr(handler, "content_type"):
                    response_headers["Content-Type"] = handler.content_type
                elif hasattr(resp, "content_type"):
                    response_headers["Content-Type"] = resp.content_type
                else:
                    raise RuntimeError(
                        "No content-type provided for raw response"
                    )
                if hasattr(handler, "headers"):
                    if isinstance(handler.headers, dicts):
                        response_headers.update(handler.headers)
                    else:
                        response_headers.update(dict(handler.headers))
                result = resp
            else:
                if (
                    not is_generator
                    and content_type == "application/json-nullstream"
                ):
                    content_type = "application/json"
                if content_type == "application/json-nullstream":
                    result = resp
                    response_headers["Content-Type"] = content_type
                elif is_generator:
                    result = resp
                else:
                    if resp is not None:
                        response[handler_name] = resp
                    if content_type == "application/bson":
                        result = bson.dumps(response)
                        response_headers["Content-Type"] = content_type
                    elif content_type == "internal/call":
                        result = response
                    elif getattr(resp, "data_type", None) == "json":
                        result = '{"status":200,"status_message":"OK","'
                        result += handler_name
                        result += '":'
                        if isinstance(resp, bytes):
                            result += resp.decode("utf-8")
                        elif isinstance(resp, str):
                            result += resp
                        else:
                            result += str(resp)
                        result += "}"
                    else:
                        result = json2.dumps(response, nice=self.niceOutput)

            if jsonp:
                if is_generator:
                    result = [jsonp, "(", result, ")"]
                else:
                    result = f"{jsonp}({result})"
            if self.add_content_length and not is_generator:
                try:
                    log.warning(
                        "Tady se špatně počítá content-length na stringu"
                    )
                    content_length = len(result)
                except TypeError:
                    content_length = result.len
                response_headers["Content-Length"] = str(content_length)
            response_headers = list(response_headers.items())
            log.debug("Starting response with headers: %r", response_headers)
            start_response(
                self._response_code(getattr(result, "http_code", 200)),
                response_headers,
            )

            if getattr(handler, "raw_response", False):
                if isinstance(result, str):
                    yield result.encode("utf8")
                elif isinstance(result, bytes):
                    yield result
                else:
                    for part in result:
                        if isinstance(part, bytes):
                            pass
                        elif not isinstance(part, str):
                            log.warning(
                                "Response date is not string but %s it can "
                                "lead to errors",
                                type(part),
                            )
                        else:
                            part = part.encode("utf8")
                        # endif
                        yield part
                return
            if is_generator:
                if content_type == "internal/call":
                    for x in result:
                        if isinstance(x, bytes):
                            yield x
                        else:
                            yield x.encode("utf8")
                        # endif
                elif content_type == "application/json-nullstream":
                    for x in result:
                        yield json2.dumps(x, nice=self.niceOutput).encode(
                            "utf8"
                        )
                        yield b"\x00"
                    # endfor
                    if self.nullstream_append_terminator:
                        yield b"\x03"
                else:
                    r_base = """{
                        "status":200,
                        "status_message":"OK",
                        "%s":[""" % (
                        handler_name,
                    )
                    # first je tady aby se ošetřila chyba genrování protože
                    #   generátor se vyhodnocuje až při první iteraci
                    try:
                        first = next(result)
                        r_base += json2.dumps2(first)
                    except StopIteration:
                        pass

                    yield r_base.encode("utf8")
                    for part in result:
                        r_base += json2.dumps2(first)
                        yield b"," + json2.dumps2(part).encode("utf8")
                    yield "]}".encode("utf8")
                return
            if isinstance(result, bytes):
                yield result
                return
            yield result.encode("utf8")
            return
        except BaseRPCException as exc:
            if isinstance(exc, ProxyException):
                exc_log.info(
                    "ExcResponse on %r: %s: %r: %r",
                    handler_name,
                    exc,
                    exc.status,
                    exc.status_message,
                )
            else:
                exc_log.info("ExcResponse on %r: %s", handler_name, exc)
            if getattr(exc, "silent", False):
                exc_log.debug(
                    "ExcResponse on %r: %s", handler_name, exc, exc_info=True
                )
            response_headers = self.response_headers[:]
            response = {
                "status": exc.status,
                "status_message": exc.status_message,
            }
            if hasattr(exc, "response_data"):
                response[handler_name] = exc.response_data
            # endif
            if hasattr(exc, "headers"):
                for header, value in exc.headers.items():
                    response_headers.append((header, value))
            if hasattr(exc, "content"):
                response = exc.content
            else:
                if content_type == "application/bson":
                    response_headers[0] = ("Content-Type", content_type)
                    response = bson.dumps(response)
                elif content_type == "internal/call":
                    pass  # internal neencodujeme do stringu
                else:
                    response = json2.dumps(response)
                if jsonp:
                    response = f"{jsonp}({response})"

            if self.add_content_length:
                log.warning("Tady se špatně počítá content-length na stringu")
                response_headers.append(("Content-Length", str(len(response))))
            start_response(
                self._response_code(200 if reply_ok_only else exc.http_status),
                response_headers,
            )
            yield response.encode("utf8")
            return
        except NotImplementedError as exc:
            exc_log.info("ExcResponse: %r (%s)", exc, type(exc))
            exc_log.debug("Not implemented raised from: %r", exc, exc_info=True)
            response_headers = self.response_headers[:]
            response = {
                "status": 501,
                "status_message": "This functionality is not implemented",
            }
            if content_type == "application/bson":
                response = bson.dumps(response)
                response_headers[0] = ("Content-Type", content_type)
            elif content_type == "internal/call":
                pass  # internal neencodujeme do stringu
            else:
                response = json2.dumps(response)
            # endif
            if jsonp:
                response = f"{jsonp}({response})"

            if self.add_content_length:
                log.warning("Tady se špatně počítá content-length na stringu")
                response_headers.append(("Content-Length", str(len(response))))
            start_response(
                self._response_code(200 if reply_ok_only else 501),
                response_headers,
            )
            yield response.encode("utf8")
            return

        except GeneratorExit:
            # response_headers = self.response_headers[:]
            # exception_str = traceback.format_exc()
            log.exception(
                "Exception GeneratorExit while calling %s", handler_name
            )
            # response_headers.append(
            #     ("Content-Length", str(len(response)))
            # )
            # start_response(
            #     self._response_code(200 if reply_ok_only else 500),
            #     response_headers
            # )
            raise
        #             return

        except BaseException:
            response_headers = self.response_headers[:]
            if self.environment == "prod":
                exception_str = f"Exception in {handler_name}"
                log.exception(handler_name)
            else:
                exception_str = traceback.format_exc()
            response = {
                "status": 500,
                "status_message": exception_str,
            }
            log.exception("Exception while calling %s", handler_name)
            if content_type == "application/bson":
                response = bson.dumps(response)
                response_headers[0] = ("Content-Type", content_type)
            elif content_type == "internal/call":
                pass  # internal neencodujeme do stringu
            else:
                response = json2.dumps(response)
            # endif
            if jsonp:
                response = f"{jsonp}({response})"

            if self.add_content_length:
                log.warning("Tady se špatně počítá content-length na stringu")
                response_headers.append(("Content-Length", str(len(response))))
            start_response(
                self._response_code(200 if reply_ok_only else 500),
                response_headers,
            )
            yield response.encode("utf8")
            return

    def _response_code(self, code):
        if sys.version_info[0] == 3 and sys.version_info[1] <= 4:
            code_name = HTTPStatus.responses[code]
        else:
            code_name = HTTPStatus(code).name

        rc = f"{int(code)} {code_name}"
        return rc

    def _get_pref_resp_ctype(self, c_type, environment):
        if c_type == "internal/call":
            return c_type
        acc = environment.get("HTTP_ACCEPT", None)
        if not acc:
            return c_type
        # endif
        acc = acc.split(",")
        acc = [a.split(";", 1)[0].strip().lower() for a in acc]
        if "*/*" in acc:
            acc.remove("*/*")
        if not acc:
            return c_type
        if "application/json-nullstream" in acc:
            return "application/json-nullstream"
        if self.bson and "application/bson" in acc:
            return "application/bson"
        if "application/json" in acc:
            return "application/json"
        log.warning("Nepodoprovane accepty:%r", environment.get("HTTP_ACCEPT"))
        return c_type
        # endif

    # enddef

    def _get_post_data(self, environment):
        try:
            request_body_size = int(environment.get("CONTENT_LENGTH", 0))
        except ValueError:
            request_body_size = 0

        if request_body_size:
            request_body = environment["wsgi.input"].read(request_body_size)
            if request_body:
                return request_body
        return None

    def _query_params(self, environment, get_raw_post_data):
        params = self._http_get_query_params(environment)

        content_type = "application/x-www-form-urlencoded"
        # charset = "utf-8"
        ct = environment.get("CONTENT_TYPE", content_type)
        ct = tuple(i.strip() for i in ct.split(";") if i.strip())
        ct_len = len(ct)
        if ct_len > 0:
            content_type = ct[0]
        # if ct_len > 1:
        #     charset = ct[1]

        # log.debug(
        #     "content_type:%s charset:%s", content_type, charset)
        raw_post_data = None
        if content_type in ("application/x-www-form-urlencoded", "text/plain"):
            if get_raw_post_data:
                raw_post_data = self._get_post_data(environment)
                if not raw_post_data:
                    raise MissingArgument("", "Empty POST data")
            else:
                params.update(self._form_encoded_params(environment))
            # endif
        elif content_type in ("application/json", "text/json", "text/plain"):
            if get_raw_post_data:
                raw_post_data = self._get_post_data(environment)
                if not raw_post_data:
                    raise MissingArgument("", "Empty POST data")
            else:
                params.update(self._json_params(environment))
        elif content_type == "application/bson":
            if not self.bson:
                raise RuntimeError("bson not enabled")
            params.update(self._bson_params(environment))
        elif content_type == "internal/call":
            params.update(environment["params"])
        else:
            if get_raw_post_data:
                raw_post_data = self._get_post_data(environment)
            else:
                raise RuntimeError(
                    f"Unknown input content-type:{content_type!r}"
                )
        # endif

        return (
            RPC._normalizeParams(params),
            self._get_pref_resp_ctype(content_type, environment),
            raw_post_data,
        )

    @staticmethod
    # pylint: disable=invalid-name
    def _normalizeParams(params):
        """
        Parametry chci vzdycky tak aby klic v slovniku byl str
        """

        p_type = type(params)
        if p_type == dict:
            return dict(
                (RPC._normalizeParams(n), RPC._normalizeParams(v))
                for n, v in params.items()
            )
        if p_type == list:
            return [RPC._normalizeParams(i) for i in params]
        if p_type == tuple:
            return tuple(RPC._normalizeParams(i) for i in params)
        if p_type == set:
            return set(RPC._normalizeParams(i) for i in params)
        return params

    # enddef

    def _http_get_query_params(self, env):
        query_params = parse_qs(env["QUERY_STRING"])
        query_params = {k: unquote(v[0]) for k, v in query_params.items()}
        return query_params

    def _form_encoded_params(self, env):
        try:
            request_body_size = int(env.get("CONTENT_LENGTH", 0))
        except ValueError:
            request_body_size = 0
        # endtry
        if request_body_size:
            request_body = env["wsgi.input"].read(request_body_size)
            if not request_body:
                raise MissingArgument("", "Empty POST data")
            # endif
            post_params = parse_qs(request_body)
        else:
            post_params = {}
        # endif
        post_params = {
            k.decode("utf8"): (
                unquote(v[0].decode("utf8")) if v[0] is not None else v[0]
            )
            for k, v in post_params.items()
        }
        return post_params

    # enddef

    def _json_params(self, env):
        try:
            request_body_size = int(env.get("CONTENT_LENGTH", 0))
        except ValueError:
            request_body_size = 0
        if request_body_size:
            request_body = env["wsgi.input"].read(request_body_size)
            if not request_body:
                raise MissingArgument("Empty POST data", "")
            try:
                json_params = json2.loads(request_body.decode("utf8"))
            except ValueError:
                log.error("error loading data: %r", request_body)
                raise
            if not isinstance(json_params, dict):
                raise MissingArgument(
                    f"- Data not dict ({json_params!r})", type(json_params)
                )
            # endif
            return json_params
        return {}

    def _bson_params(self, env):
        try:
            request_body_size = int(env.get("CONTENT_LENGTH", 0))
        except ValueError:
            request_body_size = 0
        if request_body_size:
            request_body = env["wsgi.input"].read(request_body_size)
            if not request_body:
                raise MissingArgument("", "Empty POST data")
            bson_params = bson.loads(request_body)
            # TODO raisnout, pokud to neni dict
            return bson_params
        return {}

    def _dispatch(self, path):
        path = path.replace("__", "/")
        if path in self._methods:
            return self._methods[path]
        if self._default_method:
            return self._default_method
        raise MethodNotFound(path)

    def _list_methods(self):
        response = sorted(self._methods.keys())
        return response

    def _methods_info(self):

        resp_methods = []

        resp = {
            "methods": resp_methods,
            "base_url": self.app_base_url,
            "app_name": self.app_name,
            "version": self.app_version,
        }
        methods = sorted(self._methods.keys())
        for method_name in methods:
            if method_name in (
                self._internal_methods
                + [
                    "",
                ]
            ):
                continue
            # endif
            resp_methods.append(self._methods_info_method(method_name))
        # endfor

        return resp

    def _methods_info_method(self, method):
        resp_method = {}
        resp_method["name"] = method
        resp_method["description"] = None
        resp_method["full_description"] = None
        resp_method["input"] = None
        resp_method["raises"] = []
        resp_method["output"] = None

        mth = None
        if method in self._methods:
            mth = self._methods[method]
        elif self._default_method:
            mth = self._default_method
        else:
            raise MethodNotFound(method)

        def checker_info(attr, checker):
            if checker.args:
                log.critical("Processing %r (%r)", attr, checker)
                log.critical("ARGS: %r", repr(checker.args))
                log.critical("KWARGS: %r", repr(checker.kwargs))
            # endif

            ret = {
                "optional": checker.optional,
                "description": getattr(checker, "_descr", None),
                "checker": checker.__class__.__name__,
                "json_type": checker.json_type,
                "default": None,
            }
            if checker._default is not UNDEFINED:
                ret["default"] = checker._default

            if isinstance(checker, ListChkr):
                ret["type"] = "list"
                if checker.args:
                    for value in checker.args:
                        if not isinstance(value, Checker):
                            continue
                        # endif
                        ret["items"] = checker_info(None, value)
                return ret
            # endif list

            if isinstance(checker, EnumChkr):
                ret["type"] = "string"
            else:
                ret["type"] = checker.json_type
            if ret["type"] == "null":
                ret["type"] = None

            ret["name"] = attr
            ret["data"] = {"extra": {}, "args": []}

            if isinstance(checker, DictChkr):
                ret["type"] = "object"
                ret["properties"] = {}
                if checker.kwargs:
                    for key, value in checker.kwargs.items():
                        if isinstance(value, Checker):
                            ret["properties"][key] = checker_info(key, value)
                        # endif
                    # endfor
                # endif

            if isinstance(checker, EnumChkr):
                ret["enum_values"] = tuple(checker.options)

            if checker.kwargs:
                for key, value in checker.kwargs.items():
                    if isinstance(value, Checker):
                        continue
                    # endif
                    ret["data"]["extra"][key] = value
                # endfor
            # endif
            return ret

        # enddef

        # short info
        if mth.summary:
            resp_method["description"] = mth.summary
        elif mth.__doc__:
            # Convert tabs to spaces (following the normal Python rules)
            # and split into a list of lines:
            lines = mth.__doc__.expandtabs().splitlines()
            # Determine minimum indentation (first line doesn't count):
            indent = 999
            for line in lines[1:]:
                stripped = line.lstrip()
                if stripped:
                    indent = min(indent, len(line) - len(stripped))
            # Remove indentation (first line is special):
            trimmed = [lines[0].strip()]
            if indent < 999:
                for line in lines[1:]:
                    trimmed.append(line[indent:].rstrip())
            # Strip off trailing and leading blank lines:
            while trimmed and not trimmed[-1]:
                trimmed.pop()
            while trimmed and not trimmed[0]:
                trimmed.pop(0)
            trimmed_used = []
            for line in trimmed:
                if line:
                    trimmed_used.append(line)
                break
            resp_method["description"] = " ".join(trimmed_used)

        # full info
        if mth.md_help:
            resp_method["full_description"] = mth.md_help
        elif mth.__doc__:
            doc = mth.__doc__
            resp_method["full_description"] = doc

        dict_i_res = {}
        if hasattr(mth, "deco_kwargs"):
            for key, value in mth.deco_kwargs.items():
                if not isinstance(value, Checker):
                    continue
                if hasattr(value, "_doc_hidden") and value._doc_hidden is True:
                    continue
                c_i = checker_info(key, value)
                dict_i_res[key] = c_i
            # endfor
        if dict_i_res:
            resp_method["input"] = {"type": "object", "properties": dict_i_res}
        else:
            list_i_res = None
            if hasattr(mth, "deco_args"):
                for value in mth.deco_args or []:
                    if not isinstance(value, Checker):
                        continue
                    # endif
                    if (
                        hasattr(value, "_doc_hidden")
                        and value._doc_hidden is True
                    ):
                        continue
                    list_i_res = checker_info("", value)
                    break
                # endfor
            if list_i_res:
                resp_method["input"] = list_i_res

        for key, value in (
            ("status", IntChkr()),
            ("status_message", StrChkr()),
            (
                method,
                (
                    mth.response_format or AnyChkr(optional=True)
                    if hasattr(mth, "response_format")
                    else AnyChkr(optional=True)
                ),
            ),
        ):
            if not isinstance(value, Checker):
                continue
            # endif
            if hasattr(value, "_doc_hidden") and value._doc_hidden is True:
                continue
            c_i = checker_info(key, value)
            resp_method["output"] = c_i
        # endfor

        if hasattr(mth, "raises") and mth.raises:
            for e in mth.raises:
                descr = e.__doc__ or None
                if descr:
                    descr = descr.strip()
                resp_method["raises"].append(
                    {
                        "name": e.__name__,
                        "http_status": e.http_status,
                        "status": e.status,
                        "description": descr,
                    }
                )
            resp_method["raises"] = sorted(
                resp_method["raises"], key=lambda x: x["http_status"]
            )

        # Vypis other_checkers
        resp_method["other_checkers"] = []
        if hasattr(mth, "other_checkers") and mth.other_checkers:
            for _, v in mth.other_checkers or []:
                if hasattr(v, "_doc_hidden") and v._doc_hidden is True:
                    continue
                c_i = checker_info(None, v or AnyChkr())
                resp_method["other_checkers"] = c_i
            # endfor

        return resp_method

    def _openapi_v3(self):
        return self.docs_generator.openapi_v3()

    # enddef
    _openapi_v3.raw_response = True

    def _method_index(self):
        return self.docs_generator.index()

    _method_index.raw_response = True
    _method_index.content_type = "text/html; charset=utf8"

    def _method_info(self, method=None, **kwargs):
        devel = "devel" in kwargs
        if method is None:
            if "html" in kwargs:
                return self.docs_generator.method_list_html(devel=devel)

            return self.docs_generator.method_list_plain(devel=devel)

        if "html" in kwargs:
            return self.docs_generator.method_info_html(
                method=method, devel=devel
            )

        return self.docs_generator.method_info_plaintext(
            method=method, devel=devel
        )

    _method_info.raw_response = True
    _method_info.content_type = "text/plain; charset=utf8"

    def _method_help(self, method=None):
        mth = None
        if method in self._methods:
            mth = self._methods[method]
        elif self._default_method:
            mth = self._default_method
        else:
            raise MethodNotFound(method)
        response = mth.__doc__
        if response is None:
            raise MethodHelpNotFound(method)
        response = response.strip()
        if not response:
            raise MethodHelpNotFound(method)
        return response

    def _method_status(self, check=None):
        ret = {
            "running": 1,
            "introspective": int(self.enable_introspective),
            "bson_support": self.bson,
        }
        if check in (None, "uwsgi"):
            if uwsgi is None:
                ret["uwsgi_status"] = "Uwsgi module not available"
            else:
                try:
                    ret["uwsgi.status"] = 1
                    ret["uwsgi.numproc"] = uwsgi.numproc
                    ret["uwsgi.buffer_size"] = uwsgi.buffer_size
                    ret["uwsgi.started_on"] = strftime(
                        "%Y-%m-%d %H:%M:%S", localtime(uwsgi.started_on)
                    )
                    ret["uwsgi.run_time"] = int(time() - uwsgi.started_on)
                    ret["uwsgi.version"] = uwsgi.version
                    ret["uwsgi.total_requests"] = uwsgi.total_requests()
                    ws = uwsgi.workers()
                    #            ret["uwsgi.workers"] = ws
                    working = 0
                    r_cnt = 0
                    r_time = 0
                    for one in ws:
                        if one["status"] != b"idle":
                            working += 1
                        # endif
                        r_cnt += one["requests"]
                        r_time += one["requests"] * one["avg_rt"]
                    # endfor
                    working = (working / float(len(ws))) * 100
                    ret["uwsgi.workload"] = working
                    ret["uwsgi.avg_rt"] = (
                        (r_time / r_cnt) / 1000.0 if r_cnt else 0
                    )
                except BaseException as exc:
                    ret["uwsgi_status"] = f"{exc.__class__.__name__}:{exc}"
                # endtry

        # check server status:
        ret["servers"] = []
        for server in self._servers:
            one = {}
            if check in (None, "postgre"):
                if hasattr(server, "postgre"):
                    try:
                        server._method_alive_postgre()
                        one["psql"] = 1
                        one["psql_message"] = "OK"
                    except BaseException as exc:
                        one["psql"] = 0
                        one["psql_message"] = f"{exc.__class__.__name__}:{exc}"
                    # endtry
                # endif
            if check in (None, "mysql"):
                if hasattr(server, "mysql"):
                    try:
                        server._method_alive_mysql()
                        one["mysql"] = 1
                        one["mysql_message"] = "OK"
                    except BaseException as exc:
                        one["mysql"] = 0
                        one["mysql_message"] = f"{exc.__class__.__name__}:{exc}"
                    # endtry
                # endif

            if check in (None, "cache"):
                if hasattr(server, "cache"):
                    try:
                        data = int(time())
                        server.cache.set("check", data)
                        res = server.cache.get("check")
                        if isinstance(res, (str, bytes)) and res.isdigit():
                            res = int(res)
                        if res != data:
                            one["cache"] = 0
                            one["cache_message"] = (
                                f"SET don't match GET ('{data}'!='{res}')"
                            )
                        else:
                            one["cache"] = 1
                            one["cache_message"] = "OK"
                    except BaseException as exc:
                        one["cache"] = 0
                        one["cache_message"] = f"{exc.__class__.__name__}:{exc}"
                    # endtry
                # endif

            if check in (None, "redis"):
                if hasattr(server, "redis"):
                    try:
                        data = int(time())
                        server.redis.set("check", data)
                        res = int(server.redis.get("check"))
                        if res != data:
                            one["redis"] = 0
                            one["redis_message"] = (
                                f"SET don't match GET ('{data}'!='{res}')"
                            )
                        else:
                            one["redis"] = 1
                            one["redis_message"] = "OK"
                    except BaseException as exc:
                        one["redis"] = 0
                        one["redis_message"] = f"{exc.__class__.__name__}:{exc}"
                    # endtry
                # endif
            if check in (None, "duality_master"):
                is_duality_master = getattr(server, "duality_master", None)

                if is_duality_master is None:
                    is_duality_master = self.duality_master
                is_duality_master = 1 if is_duality_master is True else 0
                one["duality_master"] = is_duality_master
                one["duality_master_message"] = (
                    "Is master" if is_duality_master == 1 else "Is not master"
                )
            # endif
            if check and check.startswith("pipver_"):
                pip_ret = {
                    "version": None,
                    "status": "unknown",
                    "message": None,
                }
                one["pipver"] = pip_ret
                try:
                    m_name = check[7:]
                    m_mod = __import__(m_name, fromlist=[""])
                    if hasattr(m_mod, "__version__"):
                        pip_ret["version"] = m_mod.__version__
                        pip_ret["status"] = "ok"
                    else:
                        pip_ret["status"] = "error"
                        pip_ret["message"] = "no module version attribute"
                except ModuleNotFoundError:
                    pip_ret["status"] = "error"
                    pip_ret["message"] = "module not found"
                except BaseException as e:
                    pip_ret["status"] = "error"
                    pip_ret["message"] = str(e)
                # endtry
            # endif
            if hasattr(server, "stat"):
                try:
                    one.update(server.stat(check=check))
                except BaseException as exc:
                    one["server_stat"] = 0
                    one["server_stat_message"] = (
                        f"{exc.__class__.__name__}:{exc}"
                    )
                # endtry
            # endif
            ret["servers"].append(one)
        # endfor

        return ret

    # pylint: disable=unused-argument
    def _method_alive(self, check=None):
        ret = True
        if ret:
            if uwsgi is None:
                ret &= False
            else:
                try:
                    ws = uwsgi.workers()
                    assert len(ws) > 0, "no workers"
                except BaseException:
                    log.exception("Error")
                    ret &= False
                # endtry

        # check server status:
        # pylint: disable=too-many-nested-blocks
        if ret:
            for server in self._servers:
                if ret:
                    if hasattr(server, "postgre"):
                        try:
                            server._method_alive_postgre()
                        except BaseException:
                            log.exception("Postgre connection error")
                            ret &= False
                        # endtry
                    # endif
                if ret:
                    if hasattr(server, "mysql"):
                        try:
                            server._method_alive_mysql()
                        except BaseException:
                            log.exception("Mysql connection error")
                        # endtry
                    # endif

                if ret:
                    if hasattr(server, "cache"):
                        try:
                            data = int(time())
                            server.cache.set("check", data)
                            res = server.cache.get("check")
                            if isinstance(res, (str, bytes)) and res.isdigit():
                                res = int(res)
                            if res != data:
                                ret &= False
                        except BaseException:
                            log.exception("Cache connection error")
                            ret &= False
                        # endtry
                    # endif

                if ret:
                    if hasattr(server, "redis"):
                        try:
                            data = int(time())
                            server.redis.set("check", data)
                            res = int(server.redis.get("check"))
                            if res != data:
                                ret &= False
                        except BaseException:
                            log.exception("Redis connection error")
                        # endtry
                    # endif
                if hasattr(server, "alive"):
                    try:
                        ret &= server.alive()
                    except BaseException:
                        log.exception("Custom method alive check exception")
                        ret &= False
                    # endtry
                # endif
            # endfor
        # endif

        return 1 if ret else 0

    def _multicall(self, methods=None, **kwargs):
        """
        Multicall ... volani vice metod najednou. Metody se zavolaji
        asynchrone (pro rychlost), ale vrati se ve spravnem poradi

        Params:
            list methods
                str     [0]     method_name
                dict    [1]     call argumets
        Result:
            list    methods
                dict    result
        """

        async def caller(args):
            method, params = args
            log.debug("Multicall:%r, %r", method, params)
            # TODO tady je potřeba backportovat nějaké fce z self.__call__
            #   kvůli vyjímkám bohužel nejde volat self.__call__ -> ten yield
            #   co tam to rozštípe
            handler = self._dispatch(method)
            try:
                res = handler(**params)
            except BaseRPCException as exc:
                return {
                    "status": exc.status,
                    "status_message": exc.status_message,
                }
            except BaseException:
                return {"status": 500, "status_message": traceback.format_exc()}
            is_generator = isinstance(res, tuple(generatorTypes))
            if is_generator:
                try:
                    res = list(res)
                except BaseException:
                    log.critical("Nedotažený error", exc_info=True)
                    return {
                        "status": 500,
                        "status_message": traceback.format_exc(),
                    }
                # endtry
            # endif
            return res

        # enddef

        async def main():
            ret = await asyncio.gather(*[caller(m) for m in methods])
            return ret

        call_results = asyncio.run(main())
        return call_results

    # enddef

    def _method_sig(self, method=None):
        return {
            k: repr(v)
            for k, v in self._methods[method].deco_kwargs.items()
            if hasattr(v, "_doc_hidden") and v._doc_hidden is not True
        }

    def _method_sig2(self, method=None):
        # lepsi pristup nez v method_info. Umoznuje zpetnou kompatibilitu, tedy
        # ze checker nemusi byt Checkr, ale obecny typ
        def checker_info(checker, keyword=None):
            ch = {
                "optional": getattr(checker, "optional", False),
                "name": checker.__class__.__name__,
                "keyword": keyword,
                "values": [],
            }
            if isinstance(checker, Checker):
                if hasattr(checker._descr):
                    ch["description"] = checker._descr
                if checker.args and len(checker.args):
                    ch["values"] = map(checker_info, checker.args)

                if checker.kwargs:
                    ch["attributes"] = [
                        checker_info(val, keyword=key)
                        for key, val in checker.kwargs.items()
                    ]
            else:
                ch["values"] = [keyword]

        return [
            checker_info(val, keyword=key)
            for key, val in self._methods[method].deco_kwargs.items()
            if hasattr(val, "_doc_hidden") and val._doc_hidden is not True
        ]

    def _method_json_schema(self, method=None):
        base = DictChkr(**self._methods[method].deco_kwargs)
        schema = base._json_schema()
        schema["$schema"] = "http://json-schema.org/schema#"
        return schema

    def _method_response_json_schema(self, method=None):
        params = {
            method: self._methods[method].response_format or AnyChkr(),
            "status": IntChkr(),
            "status_message": StrChkr(),
        }
        base = DictChkr(**params)

        schema = base._json_schema()
        schema["$schema"] = "http://json-schema.org/schema#"
        return schema

    def do_postfork(self, *args, **kwargs):
        for server in self._servers:
            server.do_postfork(*args, **kwargs)

    def at_exit(self):
        for server in self._servers:
            server.do_atexit()

    def get_server(self, server_name):
        return Server.other_servers[server_name]

    def add_server(self, server, server_name=None):
        if server_name is None:
            server_name = server.__class__.__name__
        Server.other_servers[server_name] = server
        session_methods = []
        for member_name, member in inspect.getmembers(server):
            if member_name[0] != "_":
                if hasattr(member, "exposed") and member.exposed is True:
                    if hasattr(member, "dispatch"):  # jine nazvy metod
                        dispatch_name = member.dispatch
                    else:
                        dispatch_name = member_name

                    if hasattr(member, "prefix") and member.prefix:
                        dispatch_name = member.prerfix + "/" + dispatch_name

                    server_prefix = getattr(server, "prefix", None)
                    if server_prefix:
                        dispatch_name = server_prefix + "/" + dispatch_name

                    self._methods[dispatch_name] = member
                    # defaultni metoda
                    if hasattr(member, "default") and member.default is True:
                        self._default_method = member
                if hasattr(member, "session"):
                    session_methods.append(member_name)
        if session_methods:
            new_blacklist = set(self._methods.keys()) - set(session_methods)
            self.session_blacklist += list(new_blacklist)
        # endif
        self._servers.append(server)

    # enddef


# endclass
