import unicodedata


def strip_accents(string):
    """
    zahodi diakritiku z pismen
    """
    return "".join(
        [
            c
            for c in unicodedata.normalize("NFD", string)
            if unicodedata.category(c) != "Mn"
        ]
    )


try:
    # pylint: disable=wildcard-import,unused-wildcard-import
    from sms_services import *  # noqa F403
except ImportError:
    import logging
    from xml.dom import minidom
    import urllib

    REQ_TEMPLATE = """<?xml version="1.0" encoding="UTF-8" ?>
    <messages>
        <auth>
            <name>%s</name>
            <password>%s</password>
        </auth>
    </messages>"""

    # pylint: disable=invalid-name
    class SMSPort:
        code2human = {
            # Registrace
            10: (
                "Registrace telefonního čísla na daném projektu proběhla "
                "v pořádku"
            ),
            11: (
                "Opětovná registrace telefonního čísla na daném projektu "
                "proběhla v pořádku"
            ),
            # Doručenky
            20: "SMS nenalezena",
            21: "SMS neodeslána",
            22: "SMS odeslána",
            23: "SMS doručena",
            24: "SMS nedoručena",
            25: "nepodařilo se získat status SMS",
            # Autentifikace
            100: "Nepodařila se autentifikace (username, password, ip)",
            # Interní chyby
            200: "DB problém",
            # Statusy požadavků
            0: "SMS je v pořádku a připravena k odeslání k operátorovi",
            1: (
                "Nespecifikovaná chyba (xml není validní, xml neobsahuje "
                "povinné elementy, …)"
            ),
            300: (
                "Minimálně jedna z operací nemá přidělen unikátní kód "
                "(balík byl zahozen)"
            ),
            310: (
                "SMS nemá unikátní kód (všechny SMS s tímto kódem byli "
                "odmítnuty)"
            ),
            320: "SMS nemá klíčové slovo",
            321: "Klíčové slovo dané SMS není povoleno",
            330: "Sender není specifikován",
            331: "Sender není povolen",
            332: "MO PR SMS není povoleno",
            333: "MT PR SMS není povoleno",
            334: "MT PR SMS denní limit byl překročen",
            335: "MT PR SMS maximální limit byl překročen",
            336: "Geografické číslo není povoleno",
            337: "MT SK není povoleno",
            338: "Shortkódy nejsou povoleny",
            339: "SMS má neznámý sender",
            340: "Nespecifikován typ SMS",
            341: "SMS je příliš dlouhá",
            342: "Příliš mnoho částí (max 10)",
            343: "Nesprávné číslo příjemce/odesílatele",
            350: "Nezadaný recipient nebo špatný formát",
            # Chyby XML
            701: "XML není předáno",
            702: "XML není čitelné",
            703: "Není známa metoda nebo se nejedná o POST",
            705: "XML není validní",
        }

        # pylint: disable=too-many-arguments
        def __init__(
            self,
            config=None,
            section="smsport",
            username=None,
            password=None,
            url=None,
        ):
            """
            Trida na komunikaci se SMSPortem
            """
            if config is not None:
                self._username = config.get(section, "username")
                self._password = config.get(section, "password")
                self._url = config.get(section, "url")
            else:
                self._username = username
                self._password = password
                self._url = url
            # endif

        def _fix_text(self, text):
            text = strip_accents(text)
            return text

        # pylint: disable=too-many-arguments,too-many-locals
        def send_message(
            self,
            text,
            sender,
            validity,
            dlr,
            keyword,
            priority,
            billing,
            uniq,
            recipient,
        ):
            text = self._fix_text(text)

            templ = REQ_TEMPLATE % (self._username, self._password)
            document = minidom.parseString(templ)
            msgs = document.childNodes[0]

            protomessage = document.createElement("sms")
            protomessage.setAttribute("type", "text")
            protomessage.setAttribute("sender", str(sender))
            protomessage.setAttribute("priority", str(0 if priority else 1))
            protomessage.setAttribute("dlr", str(1 if dlr else 0))
            if validity != 0:
                protomessage.setAttribute("validity", str(validity))
            if keyword is not None:
                protomessage.setAttribute("kw", str(keyword))

            txt_gen = (
                text[i : i + 160] for i in range(0, len(text), 160)  # noqa E203
            )
            i = 0
            for txt in txt_gen:
                part = document.createElement("body")
                part.setAttribute("order", str(i))
                if billing:
                    part.setAttribute("billing", "1")
                partText = document.createTextNode(txt)
                part.appendChild(partText)
                protomessage.appendChild(part)
                i += 1

            message = protomessage.cloneNode(True)
            message.setAttribute("uniq", uniq)
            message.setAttribute("recipient", recipient)
            msgs.appendChild(message)

            logging.debug("[S] sending message: %s", document.toprettyxml())

            req = urllib.request.Request(
                self._url,
                data=document.toxml(),
                headers={"Content-Type": "text/xml"},
            )
            with urllib.request.urlopen(req) as url:
                resp = url.read()

            logging.debug("[S] response: %s", resp)

            # PARSE RESPONSE
            document = minidom.parseString(resp)
            res = []
            for code in document.getElementsByTagName("code"):
                uniq = code.getAttribute("uniq")
                val = code.firstChild.data
                res.append((uniq, val))

            return res

        def check_messages(self, uniqs):
            templ = REQ_TEMPLATE % (self._username, self._password)
            document = minidom.parseString(templ)
            msgs = document.childNodes[0]

            for uniq in uniqs:
                dlr = document.createElement("dlr")
                dlr.setAttribute("uniq", str(uniq))
                text = document.createTextNode(str(uniq))
                dlr.appendChild(text)
                msgs.appendChild(dlr)

            logging.debug("[S] sending dlr request: %s", document.toprettyxml())

            req = urllib.request.Request(
                self._url,
                data=document.toxml(),
                headers={"Content-Type": "text/xml"},
            )
            with urllib.request.urlopen(req) as url:
                resp = url.read()

            logging.debug("[S] response: %s", resp)

            # PARSE RESPONSE
            document = minidom.parseString(resp)
            res = []
            for code in document.getElementsByTagName("code"):
                uniq = code.getAttribute("uniq")
                val = code.firstChild.data
                res.append((uniq, val))

            return res
