import configparser
import logging
import logging.handlers
import sys
import os
from .decorators.deprecated import deprecated
from dateutil.parser import parse as parse_dt
try:
    from .utils import UwsgiLockingFileHandler as LockingFileHandler
except ImportError:
    from .utils import LockingFileHandler

try:
    from uwsgi import opt
except ImportError:
    opt = {}


def load_config(file_path=None, load_logging=True):
    """
    helper funkce pro nacteni konfigu
    """
    config = CfgParser()
    if file_path is None:
        file_path = os.environ.get("CONF") or opt.get("conf") or opt.get("ini")
        print(
            "*** Selecting config_path from",
            os.environ.get("CONF"), opt.get("conf"), opt.get("ini"),
            "***"
        )
        if file_path and isinstance(file_path, bytes):
            file_path = file_path.decode("utf8")
        if file_path and file_path[0] != "/":
            for pfx in (
                os.getcwd() + "/",
                "/etc/uwsgi-emperor/vassals/"
            ):
                if os.path.exists(pfx + file_path):
                    file_path = pfx + file_path
                    break
    print(f"*** Loading service from {file_path} ***")
    if file_path is not None:
        if isinstance(file_path, str):
            file_path = [file_path]
        # endifi
        print("Loading configfiles:%r" % (file_path,))
        config.read(file_path)

    other_files = config.getlist("uwsgi", "load-config-files", None)
    if other_files:
        config.read(other_files)

    # nastaveni loggingu podle cfg
    if config.has_section("logging") and load_logging is True:
        _prepare_logging(config)

    return config


def _prepare_logging(config):
    """
    Logging konfigurace:
    Level=      #co logovat
    Format=     #jak ma vypadat message default:
    %(asctime)s %(levelname)s %(name)s %(filename)s %(funcName)s : %(message)s)
    Type=       #typ loggeru bud file nebo syslog, nebo stdout nebo stderr
        (default je file)

    #pro type=file:
        File=       #soubor kam logovat nebo /dev/null (default) (neloguje)

    #pro type=syslog
        Socket=     #definuje lokalni socket
            - pokud je uveden ignoruje se host a port
        Host=       #adresa serveru kde je logger (localhost)
        Port=       #port syslogu (514)

    """
    level = config.get("logging", "level", "NOTSET")
    try:
        level = getattr(logging, level)
    except AttributeError:
        level = logging.NOTSET

    logger = logging.getLogger()
    logger.setLevel(level)
    while logger.handlers:
        logger.removeHandler(logger.handlers[0])
    # endwhile

    loggerType = config.get("logging", "type", "file").lower()

    # vyberu si logformat z konfigu

    # default logformat
    logFormat = (
        "$(asctime)s [pid:$(process)d] $(levelname)s $(name)s $(filename)s"
        " $(funcName)s : $(message)s")
    if loggerType == "syslog":
        # syslog ma jiny default
        logFormat = (
            "{$(serviceName)s} $(processName)s/$(threadName)s[$(process)d/"
            "$(thread)d] $(levelname)s: $(message)s ($(filename)s:$(funcName)s"
            ":$(lineno)d)"
        )
    # endif syslog format

    # vem si format z cfg
    logFormat = config.get("logging", "format", logFormat)

    # pokud mam v sablone i nejaky servis tak si ho zjistim a upravim sablonu
    if "$(serviceName)s" in logFormat:
        logFormat = logFormat.replace(
            "$(serviceName)s", config.get(
                "logging", "serviceName", config.get(
                    "logging", "service_name",
                    "UNKNOWN_SERVICE"
                )
            ))
    # endif

    # a ted nahradim $ za % protoze % neprojde pres ConfigParser, ale logger
    # tam potrebuje % ;-)
    logFormat = logFormat.replace("$", "%")

    # a ted ten format nahradim
    logFormater = logging.Formatter(logFormat)

    # vytvorim handler (endpoint pro logovani)
    handler = None

    if loggerType == "file":
        # do dev null preci neni potreba logovat, ne??
        logFile = config.get("logging", "file", "/dev/null")
        if logFile and logFile != "/dev/null":
            handler = LockingFileHandler(logFile)
        # endif //logfile!="dev/null"
    elif loggerType == "syslog":
        host = config.get("logging", "host", "localhost")
        port = config.getint("logging", "port", 514)
        socket = config.get("logging", "socket", "")
        if socket:
            handler = logging.handlers.SysLogHandler(address=socket)
        else:
            handler = logging.handlers.SysLogHandler(address=(host, port))
        # endif
    elif loggerType in ("stdout", "stderr"):
        handler = logging.handlers.StreamHandler(
            stream=getattr(sys, loggerType))
    # endif

    if handler:
        handler.setFormatter(logFormater)
        logger.addHandler(handler)
    # endif

    logger.propagate = False

    try:
        for logger in config.options("logging-levels"):
            tmp_logger = logging.getLogger(logger)
            l_v = config.get("logging-levels", logger)
            tmp_logger.setLevel(l_v)
        # endfor
    except configparser.NoSectionError:
        logging.error(
            "Chybí sekce na detailní nastavení loggerů ([logging-levels])")
    except BaseException:
        logging.exception("Chyba načítání detailních loglevelů")
    # endry

# enddef


class CfgParser:

    class NoDefault:
        pass

    __noDedefault = NoDefault()

    def __init__(self, *args, **kwargs):
        kwargs['strict'] = False
        self.cp = configparser.ConfigParser(*args, **kwargs)

    def __getattr__(self, attr):
        return getattr(self.cp, attr)

    def read(self, path):
        if isinstance(path, bytes):
            path = path.decode("utf8")  # bez toho to nejde, až od python3.7

        res = self.cp.read(path)
        return res

    @deprecated("get(section,option,default)")
    def safe_get(self, section, option, default=None):
        return self.get(section, option, default=default)
    # enddef

    def __safe_get(self, funcName, section, option, default):
        m = getattr(self.cp, funcName, None)
        if m is None:
            raise AttributeError("CfgParser has no attribute %s" % funcName)
        try:
            res = m(section, option)
        except (
            configparser.NoOptionError,
            configparser.NoSectionError
        ):
            if default == self.__noDedefault:
                raise
            res = default
        except ValueError as e:
            if default == self.__noDedefault:
                raise
            logging.warning(
                "Config error [%s]-%s: %s", section, option, e)
            res = default
        # endtry
        return res
    # enddef

    def get(
        self, section, option, default=__noDedefault, fallback=__noDedefault
    ):
        if default == self.__noDedefault:
            default = fallback
        return self.__safe_get("get", section, option, default)
    # endef

    def getboolean(self, section, option, default=__noDedefault):
        return self.__safe_get("getboolean", section, option, default)
    # endef

    getbool = getboolean

    def getfloat(self, section, option, default=__noDedefault):
        return self.__safe_get("getfloat", section, option, default)
    # endef

    def getint(
        self, section, option, default=__noDedefault, fallback=__noDedefault
    ):
        if default == self.__noDedefault:
            default = fallback
        return self.__safe_get("getint", section, option, default)
    # endef

    def getdatetime(self, section, option, default=__noDedefault):
        if default == self.__noDedefault:
            data = self.get(section, option)
        else:
            data = self.get(section, option, "__use_default__")
            if data == "__use_default__" or not data.strip():
                return default
            # endif
        # endif
        return parse_dt(data)
    # endef

    def getdate(self, section, option, default=__noDedefault):
        if default == self.__noDedefault:
            data = self.get(section, option)
        else:
            data = self.get(section, option, "__use_default__")
            if data == "__use_default__" or not data.strip():
                return default
            # endif
        # endif
        return parse_dt(data).date()

    # endef

    def getlist(self, section, option, default=__noDedefault):
        if default == self.__noDedefault:
            data = self.get(section, option)
        else:
            data = self.get(section, option, "__use_default__")
            if data == "__use_default__":
                return default
        if not data:
            return []

        return [one.strip()
                for one in data.split(",")
                if one.strip()]

    def getintlist(self, section, option, default=__noDedefault):
        return [int(i) for i in self.getlist(section, option, default)]

    def optionsdict(self, section, fun=None, key_fun=None):
        fun = fun or (lambda x: x)
        key_fun = key_fun or (lambda x: x)

        return {
            key_fun(i[0]): fun(i[1])
            for i in self.items(section)
        }
    # enddef

    def read_update(self, filenames):
        new_config = self.__class__()
        new_config.read(filenames)
        for section in new_config.sections():
            if not self.has_section(section):
                self.add_section(section)

            for option, value in new_config.items(section):
                self.set(section, option, value)

# endclass

# eof
