#!/usr/bin/python
# -*- coding: utf-8 -*-
from logging import getLogger
from contextlib import contextmanager
from .pymysql_pool import Pool as BasePool
from .pymysql_pool import pymysql


log = getLogger("jsonrpc.connectors.mysql_pool")


class Pool(BasePool):

    @contextmanager
    def get_cursor(self, cursor_class=None):
        conn = None

        def get_cur(t_conn=None):
            if not t_conn:
                t_conn = self.get_conn()
            if cursor_class:
                t_cur = t_conn.cursor(cursor_class)
            else:
                t_cur = t_conn.cursor()
            return t_conn, t_cur

        def check_cur(t_cur):
            try:
                cur.execute("SELECT TRUE")
                cur.fetchone()
            except pymysql.OperationalError as exc:
                if exc.args[0] in (2013, 2006):
                    return False
                log.exception(
                    "Unhandled OperationalError during cursor check:%r",
                    exc.args[0]
                )
                raise
            except BaseException:
                log.exception("Unhandled Error during cursor check")
                raise
            return True

        try:
            conn, cur = get_cur()
            if not check_cur(cur):
                with self.cond:
                    self.current_size -= 1
                    # remove dead connection
                    self.inuse_list.remove(conn)
                    # create_new
                    conn = self.create_conn()
                    self.unuse_list.remove(conn)
                    self.inuse_list.add(conn)
                    self.current_size += 1
                    conn, cur = get_cur(conn)
                if not check_cur(cur):
                    raise pymysql.OperationalError(
                        999, "Unable to find healthy connection"
                    )
            yield cur
        except (
            pymysql.ProgrammingError, pymysql.OperationalError,
            pymysql.InterfaceError,
            pymysql.DatabaseError
        ) as exc:
            log.warning(
                "Removing potencialy broken connection", exc_info=True
            )
            if conn:
                with self.cond:
                    self.current_size -= 1
                    self.inuse_list.remove(conn)
                    self.cond.notify_all()

                try:
                    conn.rollback()
                except BaseException:
                    raise exc
            # reset connection to one => to prevent calling self.release(conn)
            conn = None
            raise
        except BaseException as exc:
            if conn:
                try:
                    conn.rollback()
                except BaseException:
                    raise exc
            raise
        else:
            if conn:
                conn.commit()
        finally:
            if conn:
                try:
                    self.release(conn)
                except KeyError:
                    log.warning("Releasing unknown or released connection")
    # enddef

    def create_conn(self):
        """Create mysql connection by pymysql and to add unuse_list"""
        conn = pymysql.connect(
            host=self.host,
            port=self.port,
            user=self.user,
            password=self.password,
            db=self.db,
            charset=self.charset,
            cursorclass=self.cursorclass,
            autocommit=self.autocommit,
            unix_socket=self.unix_socket,
            **self.configs
        )
        cur = conn.cursor()
        cur.execute("set names %s" % (self.charset, ))
        conn.commit()
        cur.close()
        self.unuse_list.add(conn)
        return conn


def connect_mysql_pool(
    config=None, section="mysql",
    host=None, user=None, password=None, database=None, port=None,
    charset=None, pool_min_size=None, pool_max_size=None
):
    """
    helper funkce pro napojeni databaze
        Host=       #server na kterem to bezi
        Port=       #port kde to bezi
        Username=   #jmeno uzivatele
        Password=   #heslo
        Database=   #nazev databaze
    """

    if config:
        charset = config.get(section, "charset", "utf8")
        conn = Pool(
            host=config.get(
                section, "host",
                config.get(
                    section, "hostname", "localhost"
                )
            ),
            user=config.get(section, "username", "user"),
            password=config.get(section, "password", "pass"),
            db=config.get(section, "database", "db"),
            port=config.getint(section, "port", 3306),
            charset=charset,
            autocommit=False,
            min_size=config.getint(section, "min_size", 1),
            max_size=config.getint(section, "max_size", 5),
        )
        conn.config_section = section
        conn.config_object = config
    else:
        connect_args = {
            "host": host,
            "user": user,
            "password": password,
            "database": database,
            "port": port or 3306,
            "charset": charset or "utf8",
            "autocommit": False,
            "min_size": pool_min_size if pool_min_size is not None else 1,
            "max_size": pool_max_size if pool_max_size is not None else 5
        }
        conn = Pool(**connect_args)
        conn.connect_args = connect_args

    return conn
# enddef connect_db

# eof
