from collections import namedtuple
from logging import getLogger
from ..wrapper_classes import register_generator


log = getLogger("jsonrpc.connectors.postgresql")
log_d = getLogger("jsonrpc.connectors.postgresql_conn")

try:
    import psycopg2
    import psycopg2.extras
except BaseException:
    log.exception("Nemohu naimportovat python pyckage: Psycopg2")

    class psycopg2:
        class Error(Exception):
            pass

        class OperationalError(Exception):
            pass

        class extras:
            class NamedTupleCursor:
                pass

            class RealDictCursor:
                pass

            class MyNamedTupleCursor:
                pass

    pass
else:
    # pokud chceme používat na venek i kurzory musíme s nimi pracovat
    #   jako s generátory a ne všechny jsou => musíme si je virtuálně
    #   přidat mezi generátory
    register_generator(psycopg2.extras.NamedTupleCursor)
    register_generator(psycopg2.extras.RealDictCursor)
# endtry


class Record(object):
    def __init__(self, desc=None):
        if desc:
            self.__obj = namedtuple("record_i", desc)

    def _make(self, *args, **kwargs):
        obj = Record()
        obj.__obj = self.__obj._make(*args, **kwargs)
        return obj

    def __repr__(self, *args, **kwargs):
        return self.__obj.__repr__(*args, **kwargs)

    def _asdict(self, *args, **kwargs):
        return self.__obj._asdict(*args, **kwargs)

    def __getattr__(self, attr):
        return getattr(self.__obj, attr)

    def __getstate__(self):
        return self.__obj._asdict()

    def __setstate__(self, state):
        nt = namedtuple("record_s", list(state.keys()))
        self.__obj = nt._make(list(state.values()))

    def __iter__(self):
        return self.__obj.__iter__()

    def __getitem__(self, key):
        return self.__obj.__getitem__(key)

    def next(self):
        # py3: next(self.__obj) ?
        return self.__obj.next()

    def __len__(self):
        return self.__obj.__len__()


class MyNamedTupleCursor(psycopg2.extras.NamedTupleCursor):

    def _make_nt(self, namedtuple=namedtuple):
        return Record([d[0] for d in self.description or ()])


psycopg2.extras.Record = Record
psycopg2.extras.MyNamedTupleCursor = MyNamedTupleCursor

register_generator(MyNamedTupleCursor)


class WrapClass(object):
    _xconn = None
    connect_args = None

    def __init__(self, conn, c_args):
        self._xconn = conn
        self.connect_args = c_args

    def __getattr__(self, attr):
        return getattr(self._xconn, attr)

    def __setattr__(self, attr, val):
        if attr in ("_xconn", "connect_args"):
            object.__setattr__(self, attr, val)
        else:
            setattr(self._xconn, attr, val)


def connect_postgre(
    config=None, section="postgre",
    host=None, user=None, password=None, database=None, port=None,
    charset=None, readonly=False
):
    """
    """
    c_args = {
        "config": config,
        "section": section,
        "host": host,
        "user": user,
        "password": password,
        "database": database,
        "port": port,
        "charset": charset,
        "readonly": readonly
    }
    if config:
        host = config.get(
            section, "host", config.get(section, "hostname", None)
        )
        readonly = config.getboolean(section, "readonly", False)
        if readonly:
            host = config.get(
                section, "host_readonly", config.get(
                    section, "hostname_readonly", host
                )
            )

        if not host:
            raise ValueError(
                "missing 'host' in config (section %r)" % (section,)
            )
        conn_args = dict(
            host=host,
            user=(
                config.get(section, "username") or
                config.get(section, "user")
            ),
            password=config.get(section, "password"),
            database=config.get(section, "database"),
            port=config.getint(section, "port", 5432),
        )
        schm = config.get(section, "schema", None)
        if schm:
            conn_args['options'] = '-c search_path=' + schm
        try:
            conn = psycopg2.connect(**conn_args)
        except BaseException:
            log.error("Connect faied using:%r", conn_args)
            raise

        # kodovani
        conn.set_client_encoding(config.get(section, "charset", "utf8"))
        config.postgre_config_section = section
        config.config_object = config
        conn.set_session(readonly=readonly)
        log_d.debug("New connection ro=%s cfg %r", readonly, conn_args)
    else:
        conn = psycopg2.connect(
            host=host,
            user=user,
            password=password or None,
            database=database,
            port=port or 5432
        )
        # kodovani
        conn.set_client_encoding(charset or "utf8")
        conn.set_session(readonly=readonly)
        log_d.debug(
            "New connection ro=%s args (%s, %s, %s)",
            readonly, host, database, user
        )
    # endif

    # ze slovniku automaticky json
    psycopg2.extensions.register_adapter(dict, psycopg2.extras.Json)
    """
    py2 only
    if config.getboolean(section, "string_as_unicode", False):
        psycopg2.extensions.register_type(psycopg2.extensions.UNICODE)
        psycopg2.extensions.register_type(psycopg2.extensions.UNICODEARRAY)
    # endif
    """

    psycopg2.extras.register_uuid(conn_or_curs=conn)

    return WrapClass(conn, c_args)
# enddef
