#!/usr/bin/python
# -*- coding: utf-8 -*-

import functools
import logging
from copy import deepcopy

from ..utils import LockingFileHandler
from ..wrapper_classes import generatorTypes
from ..checkers import Checker, UNDEFINED, ListChkr
from ..exceptions import MissingArgument, UnknownArgument, RPCException

log = logging.getLogger("jsonrpc.decorators.method")

res_check_log = logging.getLogger("jsonrpc.response_check")

request_loglevels = {}
request_log = logging.getLogger("jsonrpc.decorators.request_log")


class ResponseCheckFailed(RPCException):
    def __init__(self, status_message):
        self.status = 550
        self.status_message = status_message


def get_request_loglevel(config, method):
    level = request_loglevels.get(method)
    if level is None:
        level = None
        if config:
            lname = "jsonrpc.decorators.request_log." + method
            for s, o in (
                ("logging-levels", lname),
                ("logging-levels", "jsonrpc.decorators.request_log"),
                ("logging", "request_logfile_level"),
                ("logging", "level"),
            ):
                level = config.get(s, o, None)
                if level:
                    break
        level = level if level is not None else "INFO"
        if level.isdigit():
            level = int(level)
        else:
            level = getattr(logging, level, None)
            if level is None or not isinstance(level, int):
                level = logging.INFO
        # endif
        request_loglevels[method] = level
        init_request_log(config)
    # endif
    return level


def init_request_log(config):
    if not getattr(request_log, "request_log_initialized", False):
        logFormat = "%(asctime)s [pid:%(process)d] %(message)s"
        logFormater = logging.Formatter(logFormat)
        log_file = None
        if config:
            log_file = config.get("logging", "request_logfile", None)

        request_log.setLevel(1)
        if not log_file:
            req_l = request_log
            while req_l:
                for h in req_l.handlers:
                    log_file = getattr(h, "baseFilename", None)
                    if log_file:
                        break
                if log_file:
                    break
                req_l = req_l.parent
            # endwhile
        # endif

        if log_file:
            handler = LockingFileHandler(log_file)
            handler.setFormatter(logFormater)
            request_log.addHandler(handler)
            request_log.propagate = False

        request_log.request_log_initialized = True
    # endif


class method(object):
    """
    Dekorátor pro zveřejnění metody na veřejném rozhraní.
    """

    def __init__(self, *deco_args, **deco_kwargs):
        self.deco_args = deco_args
        self.deco_kwargs = deco_kwargs
        self.prefix = ""

    # enddef __init__

    def __call__(self, func):
        func.exposed = True
        func.default = False
        func.response_format = self.deco_kwargs.pop("_response", None)
        func.prefix = self.deco_kwargs.pop("_prefix", None)
        func.other_checkers = self.deco_kwargs.pop("_other_checkers", [])
        func.raises = self.deco_kwargs.pop("_raises", [])

        md_help = self.deco_kwargs.pop("_md_help", None)
        if md_help and getattr(func, "md_help", None):
            log.critical(
                "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! "
                "Function %s has both md_help as attr and in method kwargs",
                func.__name__,
            )
        elif md_help:
            func.md_help = md_help
        elif not hasattr(func, "md_help"):
            func.md_help = None
        # endif

        summary = self.deco_kwargs.pop("_summary", None)
        if summary and getattr(func, "summary", None):
            log.critical(
                "Function %s has both summary as attr and in method kwargs",
                func.__name__,
            )
        elif summary:
            func.summary = summary
        elif not hasattr(func, "summary"):
            func.summary = None
        # endif

        if func.response_format:
            # base response je v případě chyby nepřítomná
            #   -> tedy vždy je root optional
            func.response_format.optional = True
        if func.other_checkers:
            for k, v in func.other_checkers:
                v.optional = True
        # endif

        # raw
        if (
            "_raw_response" in self.deco_kwargs
            and self.deco_kwargs["_raw_response"]
        ):
            log.debug("Registered raw method %s" % func.__name__)
            func.raw_response = self.deco_kwargs.get("_raw_response")
            del self.deco_kwargs["_raw_response"]
            if "_content_type" in self.deco_kwargs:
                log.debug(
                    "Forced content-type for %s to %s"
                    % (func.__name__, self.deco_kwargs["_content_type"])
                )
                func.content_type = self.deco_kwargs["_content_type"]
                del self.deco_kwargs["_content_type"]
        else:
            func.raw_response = False
        # raw_post_data
        # defaultni metoda
        if self.deco_kwargs.get("_default"):
            func.default = True
            del self.deco_kwargs["_default"]
        else:
            func.default = False

        # vlastni cesta
        if self.deco_kwargs.get("_path"):
            func.dispatch = self.deco_kwargs["_path"]
            del self.deco_kwargs["_path"]

        # chci hlavicky
        if self.deco_kwargs.get("_headers"):
            func.request_headers = True
            del self.deco_kwargs["_headers"]

        # chci potlacit nekte dotazovaci metody
        if self.deco_kwargs.get("_silent_methods"):
            func.silent_methods = self.deco_kwargs["_silent_methods"]
            del self.deco_kwargs["_silent_methods"]

        # chci raw post data
        if self.deco_kwargs.get("_raw_post_data"):
            log.debug("Registered raw post data method %s" % func.__name__)
            func.request_raw_post_data = True
            del self.deco_kwargs["_raw_post_data"]

        # chci kompletni environment
        if self.deco_kwargs.get("_raw_environment"):
            log.debug("Registered raw environment method %s" % func.__name__)
            func.request_raw_environment = True
            del self.deco_kwargs["_raw_environment"]

        # nastaveni chovani pri nedefinovanych parametrech
        func.params_check_mode = self.deco_kwargs.pop(
            "_params_check_mode", None
        )

        # nastaveni checkerů pouze pro auth
        func._auth_kwargs = self.deco_kwargs.pop("_auth_kwargs", None) or ()

        # nastaveni cached igorování:
        func.remove_cache_args = bool(
            self.deco_kwargs.pop("_remove_cache_args", False)
        )

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            for param, value in list(kwargs.items()):
                if param == "__FLUSH" and func.remove_cache_args:
                    kwargs.pop("__FLUSH")
                    log.debug("Removing __FLUSH as requested")
                elif param in self.deco_kwargs and hasattr(
                    self.deco_kwargs[param], "__call__"
                ):
                    if isinstance(self.deco_kwargs[param], Checker):
                        kwargs[param] = self.deco_kwargs[param](
                            value, param, ""
                        )
                    else:
                        kwargs[param] = self.deco_kwargs[param](value)
                    # endif
                else:
                    if (
                        param == "cursor"
                        and args
                        and (
                            hasattr(args[0], "mysql")
                            or hasattr(args[0], "psql")
                        )
                    ):
                        log.debug("Method received cursor")
                    elif param in ("__FLUSH", "cursor"):
                        # parametry ktere se pocita ze vezmou jine dekoratory
                        pass
                    else:
                        if func.params_check_mode == "strict":
                            raise UnknownArgument(func.__name__, param, value)
                        elif func.params_check_mode == "silent":
                            pass
                        else:
                            log.warning(
                                "Function %s received unknown param %s:%s"
                                % (func.__name__, param, value)
                            )
                    # endif
                # endif
            # endfor
            for rk, rv in self.deco_kwargs.items():
                if rk in func._auth_kwargs:
                    log.debug("Skiping check of Auth checker %r", rk)
                    continue
                if isinstance(rv, Checker) and rv.optional:
                    if rk not in kwargs and rv._default is not UNDEFINED:
                        kwargs[rk] = deepcopy(rv._default)
                    # endif
                    continue
                # endif
                if rk not in kwargs:
                    raise MissingArgument(func.__name__, rk)
                # endif
            # endfor
            server = args[0] if args else None
            config = getattr(server, "config", None)
            level = get_request_loglevel(config, func.__name__)
            if level <= logging.INFO:
                request_log.info(
                    "Method: %s Params: %r call started", func.__name__, kwargs
                )
            try:
                ret = func(*args, **kwargs)
            except RuntimeError as e:
                if level <= logging.WARNING:
                    request_log.warning(
                        "Method: %s Params: %r Response: RuntimeError(%r)",
                        func.__name__,
                        kwargs,
                        str(e),
                    )
                # endif
                raise
            except RPCException as e:
                if level <= logging.WARNING:
                    data = getattr(e, "response_data", None)
                    if data is not None and level > logging.DEBUG:
                        data = "muted"
                    request_log.warning(
                        "Method: %s Params: %r ResponseError: "
                        "http:%r code:%r msg:%r data:%r",
                        func.__name__,
                        kwargs,
                        e.http_status,
                        e.status,
                        e.status_message,
                        data,
                    )
                # endif
                raise
            except BaseException as e:
                if level <= logging.WARNING:
                    request_log.warning(
                        "Method: %s Params: %r Response: %s(%s)",
                        func.__name__,
                        kwargs,
                        type(e),
                        e,
                    )
                # endif
                raise
            except BaseException:
                if level <= logging.WARNING:
                    request_log.warning(
                        "Method: %s Params: %r Response: FATAL ERROR",
                        func.__name__,
                        kwargs,
                    )
                # endif
                raise

            if level <= logging.DEBUG:
                request_log.debug(
                    "Method: %s Params: %r Response: %r",
                    func.__name__,
                    kwargs,
                    ret,
                )
            elif level <= logging.INFO:
                request_log.info(
                    "Method: %s Params: %r Response: muted",
                    func.__name__,
                    kwargs,
                )
            # endif
            enable_response_check = getattr(
                server, "_enable_response_check", False
            )
            if not enable_response_check:
                res_check_log.debug("Response format check disabled")
            elif not func.response_format:
                res_check_log.debug(
                    "Response checker not defined for %s", func.__name__
                )
            elif func.raw_response:
                res_check_log.debug(
                    "Not checking response format for raw_responses"
                )
            else:
                if not isinstance(ret, tuple(generatorTypes)):
                    try:
                        ret = func.response_format(ret, func.__name__, "")
                    except Exception as e:
                        res_check_log.exception("Response not valid")
                        if enable_response_check == "critical":
                            raise ResponseCheckFailed(e.status_message)
                    else:
                        res_check_log.info("Response valid")
                else:
                    if not isinstance(func.response_format, ListChkr):
                        res_check_log.warning(
                            "Response check skipped - response "
                            "checker is not ListChkr"
                        )
                    else:
                        # I nedd wrapper
                        def checkret(iterator, chkr):
                            for i, itm in enumerate(iterator):
                                try:
                                    chkr(itm, func.__name__, f"Item nr:{i}")
                                except Exception as e:
                                    res_check_log.exception(
                                        f"Response nr:{i} not valid"
                                    )
                                    if enable_response_check == "critical":
                                        raise ResponseCheckFailed(
                                            e.status_message
                                        )
                                else:
                                    res_check_log.debug(
                                        f"Response nr:{i} valid"
                                    )
                                yield itm
                            # endfor

                        # enddef

                        ret = checkret(ret, func.response_format.checker)
                    # endif - list checker or anything else
                # endif - generator or not?
            # endif - check response?

            return ret

        # enddef wrapper
        wrapper.deco_kwargs = self.deco_kwargs
        wrapper.deco_args = self.deco_args
        wrapper._wraps = getattr(func, "_wraps", func)
        return wrapper

    # endef __call__


# enddef method

# eof
