import functools
from logging import getLogger

from types import GeneratorType
from .helpers import pymssql
from ..connectors import connect_mssql

log = getLogger("jsonrpc.decorators.mssqlcursor")


class MsSqlCursor(object):

    def __init__(self, db_attr=None):
        self._db_attr = db_attr or "mssql"
    # enddef

    def __call__(self, func):
        def generator_wrapper(cursor, generator):
            try:
                yield from generator
            except StopIteration:
                cursor.close()
                # raise StopIteration

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            cursor_reuse = "cursor" in kwargs
            close_when_done = True
            if cursor_reuse:
                log.debug("reusing cursor from args")
                cursor = kwargs.pop("cursor")
                if cursor is None:
                    import traceback
                    log.fatal("%s", traceback.format_stack())
                    log.fatal("Nepodarilo se najit cursor v kwargs")
                    # TODO kvuli zpetne kompatibilite jen loguji, ale chtelo
                    # by to:
                    # raise ProgrammingError("chybny typ kurzoru")
                # endif
            else:
                conn = getattr(args[0], self._db_attr)

                try:
                    test_cur = conn.cursor()
                    test_cur.execute("select 1 as TestColumn")
                    test_cur.close()
                except (
                    pymssql.OperationalError,
                    pymssql.InterfaceError,
                    pymssql.InternalError,
                    pymssql.DatabaseError
                ) as e:
                    log.warning(
                        "MsSql error on check:%s:%s:%s:%s" % (
                            type(e), e,
                            getattr(e, "severity", None),
                            getattr(e, "state", None),
                        )
                    )
                    if hasattr(conn, "config") and hasattr(conn, "section"):
                        conn = connect_mssql(
                            conn.config,
                            section=conn.section
                        )
                        setattr(args[0], self._db_attr, conn)
                    else:
                        log.critical(
                            "Can't reconnect database!!! I have no config "
                            "attribute on object %s" % (type(args[0])))
                        raise
                    # endif

                # endtry
                # vytvorime cursor
                cursor = conn.cursor()
            # endif
            args = list(args)
            args.insert(1, cursor)
            try:
                res = func(*args, **kwargs)
                if isinstance(res, GeneratorType):
                    close_when_done = False
                    res = generator_wrapper(cursor, res)
                # endif

            except pymssql.ProgrammingError as e:
                log.error("Programming error: %s", e)
                log.exception("Traceback")

                if not cursor_reuse:
                    # pozor i rollback nad rozbitou DB muze vyvolat vyjimku!!
                    log.debug("ROLLBACK")
                    conn.rollback()
                # endif
                raise
            except pymssql.Error as e:
                log.exception(
                    "Error while executing query: \n args: %r, kwargs: %r",
                    args, kwargs
                )
                if not cursor_reuse:
                    if conn.conn._conn.connected:
                        log.debug("ROLLBACK")
                        try:
                            conn.rollback()
                        except BaseException:
                            log.debug("Error during rollback", exc_info=True)
                            raise e
                    else:
                        log.debug("Db is not connected - skipping ROLLBACK")

                # endif
                raise
            except Exception as e:
                if not cursor_reuse:
                    if conn.conn._conn.connected:
                        log.debug("ROLLBACK")
                        try:
                            conn.rollback()
                        except BaseException:
                            log.debug("Error during rollback", exc_info=True)
                            raise e
                    else:
                        log.debug("Db is not connected - skipping ROLLBACK")
                # endif
                raise
            else:
                if not cursor_reuse and close_when_done:
                    # pozor commit muze vyvolat vyjimku pri nesplneni treba
                    # foreign keys!!
                    # log.debug("COMMIT")
                    conn.commit()
                # endif
            # endtry
            if not cursor_reuse and close_when_done:
                cursor.close()
            # endif
            return res
        wrapper._wraps = getattr(func, "_wraps", func)
        return wrapper
    # enddef
# endclass


def mssql_cursor(
    dict_result=False, db=None, db_attr=None
):
    """
    dekorator pro databazi
    """
    return MsSqlCursor(db_attr=db_attr)
# enddef cursor

# eof
