from logging import getLogger

from .doc_gen_base import DocGenBase

log = getLogger("doc_gen_openapi3")


class DocGenOpenApi3(DocGenBase):
    def after_init(self):
        self._res = {}

    def process_server(self, data):
        super().process_server(data)
        self._res.clear()
        self._prepare_header(data)
        self._process_methods(data)
        return self._res

    def _prepare_header(self, data):
        self._res["openapi"] = "3.0.0"
        self._res["info"] = {
            "version": data.get("version") or "1.0.0",
            "title": data.get("app_name") or "NonameApp",
            "license": {"name": "MIT"},
            "description": (
                data.get("opeanapi_doc_description") or "No description"
            ),
        }
        self._res["servers"] = [
            {
                "url": data.get("base_url") or "proto://host:port",
                "description": "Api url",
            }
        ]
        self._res["paths"] = {}

    def _process_methods(self, data):
        for method in data["methods"]:
            # DEBUG
            # if method["name"] not in ("cz_ported_ctu_check", ):
            #     continue
            mpath = "/" + method["name"]
            self._res["paths"][mpath] = self._process_methods_method(method)
        # endfor

    # enddef

    def _process_methods_method(self, method):
        mret = {}
        mretc = {}
        mret["post"] = mretc
        mretc["summary"] = method["description"] or "[method description]"
        mretc["description"] = (
            method["full_description"] or "[method description]"
        )

        mretc["requestBody"] = {
            "content": {
                "application/json": {
                    "schema": self._process_methods_method_schema(
                        method["input"]
                    )
                }
            }
        }

        mretc["responses"] = {}
        mretc["responses"]["200"] = {"description": "OK"}

        if "output" in method:
            mretoc = {
                "application/json": {
                    "schema": self._process_methods_method_schema(
                        method["output"]
                    )
                }
            }
            mretc["responses"]["200"]["content"] = mretoc
        # endif

        if "raises" in method:
            groups = {}
            for one in method["raises"] or []:
                http_status = one["http_status"]
                groups.setdefault(http_status, [])
                groups[http_status].append(one)
            # endfor
            for http_status, raises in groups.items():
                mretoco = {"oneOf": []}
                mretocr = {
                    "description": "Error",
                    "content": {"application/json": {"schema": mretoco}},
                }
                mretc["responses"][str(http_status)] = mretocr

                for rai in raises:
                    mretrp = {"type": "object", "properties": {}}
                    for one in (
                        ("status", "number", "Internal number code"),
                        ("description", "string", "Error description"),
                        ("name", "string", "Error internal name"),
                    ):
                        mretrp["properties"][one[0]] = {
                            "type": one[1],
                            "description": one[2],
                            "example": rai[one[0]],
                        }
                    # endfor
                    mretrp["required"] = [
                        "status",
                        "description",
                        "name",
                    ]
                    mretoco["oneOf"].append(mretrp)
                # endfor

            # endfor
        # endif
        return mret

    # endfor

    def _process_methods_method_schema(self, params):
        if not params:
            return {"type": "object"}
        return self._process_param(params)

    @classmethod
    # pylint: disable=too-many-branches
    def _process_param(cls, obj):
        obj_type = obj["type"]

        if obj_type is None:
            return {"type": "object"}

        ret = {}
        enum_values = []
        required_properties = []
        if obj.get("description"):
            ret["description"] = " ".join(obj["description"].splitlines())

        if obj_type == "object":
            ret["type"] = "object"
            ret["properties"] = {}
            for p_name, p_value in (obj.get("properties") or {}).items():
                if not p_value.get("optional") is True:
                    required_properties.append(p_name)
                ret["properties"][p_name] = cls._process_param(p_value)
            # endfor

        elif obj_type == "list":
            ret["type"] = "array"
            ret["items"] = {}
            obj_items = obj.get("items")
            if obj_items:
                ret["items"] = cls._process_param(obj_items)

        elif obj_type in ("integer", "string", "boolean", "number"):
            ret["type"] = obj_type

        else:
            raise ValueError(f"Unknown object type {obj_type}")

        if obj_type in ("integer", "number"):
            d_e = (obj.get("data") or {}).get("extra") or {}
            d_e_min = d_e.get("min")
            if d_e_min is not None:
                ret["minimum"] = d_e_min
            d_e_max = d_e.get("max")
            if d_e_max is not None:
                ret["maximum"] = d_e_max
        # endif

        # enums
        if "enum_values" in obj and obj["enum_values"]:
            enum_values = list(obj["enum_values"])
            if obj.get("optional") is True:
                ret["nullable"] = True
                enum_values.append(None)

        # default
        if obj.get("default") and obj["default"] is not None:
            ret["default"] = obj["default"]

        if required_properties:
            ret["required"] = required_properties
        if enum_values:
            ret["enum"] = enum_values
        return ret

    # enddef


# eof
