from logging import getLogger
import os
import inspect

import json2
from checkerslib import EnumChkr, IntChkr, StrChkr, AnyChkr, Checker
from checkerslib import UNDEFINED

from .exceptions import MethodNotFound, MethodHelpNotFound
from .doc_gen.doc_gen_openapi3 import DocGenOpenApi3


log = getLogger("jsonrpc.docs")

try:
    # pylint: disable=import-error
    import yaml
except ImportError:
    yaml = None
    log.warning("Yaml not available")

md_extensions = ["fenced_code", "tables"]
try:
    import markdown
except ImportError:
    markdown = None
    log.warning("Markdown not available")
else:
    md_extensions.append("jsonrpc.md_mermaid")

    try:
        from pygments.formatters import HtmlFormatter
        from markdown.extensions.codehilite import CodeHiliteExtension
    except ImportError:
        log.warning("Markdown codehilite not available")
    else:

        class CustomHtmlFormatter(
            HtmlFormatter
        ):  # pylint: disable=too-few-public-methods
            def __init__(self, lang_str="", **options):
                super().__init__(**options)
                # lang_str has the value {lang_prefix}{lang}
                # specified by the CodeHilite's options
                self.lang_str = lang_str

            def _wrap_code(self, source):
                yield 0, f'<code class="{self.lang_str}">'
                yield from source
                yield 0, "</code>"

        # endclass

        md_extensions.append(
            CodeHiliteExtension(pygments_formatter=CustomHtmlFormatter)
        )


help_trans = {
    "cs": {
        "response": "Odpověď:",
        "params": "Parametry:",
        "vyskyt": "Výskyt:",
        "popis": "Popis:",
        "raises": "Výjimky:",
        "help_not_found": "Nápověda nenalezena!",
    },
    "en": {
        "response": "Response:",
        "params": "Parameters:",
        "vyskyt": "Occurence:",
        "popis": "Description:",
        "raises": "Exceptions:",
        "help_not_found": "Help file has not been found!",
    },
}


class Str(str):
    content_type = "text/plain; charset=utf8"


class Html(str):
    content_type = "text/html; charset=utf8"


class Yaml(str):
    content_type = "text/x-yaml; charset=utf8"


class DocsGenerator:
    config = None
    _help_lang = "cs"
    _default_help_dir = __file__.rsplit("/", 1)[0] + "/static"
    _doc_dir = None
    _doc_md_file = None
    _doc_html_file = None
    _doc_html_addon_file = None
    app = None

    def __init__(self, app, config, enable_help_call=True, help_lang="cs"):
        if enable_help_call is None:
            enable_help_call = True
        if help_lang not in help_trans:
            help_lang = "cs"
        self._help_lang = help_lang
        self.config = config
        self.app = app

        help_dir = __file__.rsplit("/", 1)[0] + "/static"
        if not enable_help_call:
            help_dir += "/no_call"
        if help_lang != "cs":
            help_dir += "/" + help_lang

        self._default_help_dir = help_dir

        doc_dir = os.getcwd() + (
            "/doc" if help_lang == "cs" else "/doc/" + help_lang
        )
        if os.path.exists(doc_dir):
            self._doc_dir = doc_dir

            doc_md_file = (
                doc_dir
                + "/"
                + (
                    self.config.get("rpc", "doc_md_file", "doc.md")
                    if self.config
                    else "doc.md"
                )
            )
            if os.path.exists(doc_md_file):
                self._doc_md_file = doc_md_file
            # endif doc_md_file

            doc_html_file = (
                doc_dir
                + "/"
                + (
                    self.config.get("rpc", "doc_html_file", "help.html")
                    if self.config
                    else "help.html"
                )
            )
            if os.path.exists(doc_html_file):
                self._doc_html_file = doc_html_file
            # endif doc_html_file

            doc_html_addon_file = (
                doc_dir
                + "/"
                + (
                    self.config.get("rpc", "doc_html_addon_file", "addon.html")
                    if self.config
                    else "addon.html"
                )
            )
            if os.path.exists(doc_html_addon_file):
                self._doc_html_addon_file = doc_html_addon_file
            # endif doc_html_addon_file

        # endif

    def _get_doc_md_content(self):
        if not self._doc_md_file:
            log.debug("Custom help file in MD not found")
            return None
        with open(self._doc_md_file, encoding="utf8") as f:
            return f.read()

    def _get_doc_md_content_in_html(self):
        if markdown is None:
            log.debug("Markdown not available")
            return None
        md = self._get_doc_md_content()
        if md is None:
            log.debug("Custom help file in MD not found")
            return None
        html = markdown.markdown(
            md, extensions=md_extensions, output_format="html"
        )
        return html

    def index(self):
        log.debug("index")
        # HEADER
        with open(self._default_help_dir + "/head.html", encoding="utf8") as f:
            for i in f:
                yield i

        # CUSTOM HELP
        if self._doc_html_file:
            with open(self._doc_html_file, encoding="utf8") as f:
                for i in f:
                    yield i
        else:
            html = self._get_doc_md_content_in_html()
            if html:
                yield html
            else:
                log.debug("Custom help file not found")
                yield (
                    f"<h3>{help_trans[self._help_lang]['help_not_found']}</h3>"
                )
        # endif

        if self._doc_html_addon_file:
            with open(self._doc_html_addon_file, encoding="utf8") as f:
                for i in f:
                    yield i

        # FOOTER
        with open(self._default_help_dir + "/foot.html", encoding="utf8") as f:
            for i in f:
                yield i

    @staticmethod
    def _is_internal_method(method):
        return method.startswith("system_") or method in ("openapi_v3", "")

    def method_list_html(self, devel=False):

        d = "&devel=1" if devel else ""
        return Html(
            "<br>".join(
                f'<a href="/system_method_info?method={i}&html=1{d}">{i}</a>'
                for i in sorted(self.app._methods.keys())
                if not self._is_internal_method(i)
            )
        )

    # pylint: disable=unused-argument
    def method_list_plain(self, devel=False):
        return Str(
            "\n".join(
                i
                for i in sorted(self.app._methods.keys())
                if not self._is_internal_method(i)
            )
        )

    def method_info_html(self, method, devel=False):
        return Html(
            self._format_method_info(
                method,
                devel=devel,
                opt_start='<li class="optional">',
                opt_end_inner=False,
                opt_end="</li>",
                mand_start='<li class="mandatory">',
                mand_end="</li>",
                group_start='<ul id="group-%s">',
                group_end="</ul>",
                nl="",
                indent_str="",
                heading_start="<h3>",
                heading_end="</h3>",
                info_start="<p>",
                info_end="</p>",
                arg_descr_start='<p class="desc">',
                arg_descr_end="</p>",
                attr_name_start='<b class="type_%s">',
                attr_name_end="</b>",
                enum_info_start="<em>",
                enum_info_end="</em>",
                markdown_to_html=True,
            )
        )

    def method_info_plaintext(self, method, devel=False):
        return Str(
            self._format_method_info(
                method,
                devel=devel,
                nl="\n",
                indent_str="\t",
                opt_start="[",
                opt_end_inner=True,
                opt_end="]",
                mand_start="*",
                mand_end="",
                group_start="",
                group_end="",
                heading_start="",
                heading_end="",
                info_start="",
                info_end="",
                arg_descr_start=" - ",
                arg_descr_end="",
                attr_name_start="",
                attr_name_end="",
                enum_info_start="",
                enum_info_end="",
                markdown_to_html=False,
            )
        )

    # pylint: disable=too-many-arguments, too-many-branches,too-many-statements
    # pylint: disable=too-many-locals
    def _format_method_info(
        self,
        method,
        devel=False,
        nl="\n",
        indent_str="\t",
        opt_start="[",
        opt_end_inner=True,
        opt_end="]",
        mand_start="*",
        mand_end="",
        group_start="",
        group_end="",
        heading_start="",
        heading_end="",
        info_start="",
        info_end="",
        arg_descr_start=" - ",
        arg_descr_end="",
        attr_name_start="",
        attr_name_end="",
        enum_info_start="",
        enum_info_end="",
        markdown_to_html=False,
    ):

        trans = help_trans[self._help_lang]

        mth = None
        if method in self.app._methods:
            mth = self.app._methods[method]
        elif self.app._default_method:
            mth = self.app._default_method
        else:
            raise MethodNotFound(method)

        def checker_info(attr, checker, indent=0, indent_str=" "):
            nl_indent = indent_str * indent
            res = nl_indent
            res += mand_start if not checker.optional else opt_start
            chkr_name = checker.__class__.__name__.replace("Chkr", "").lower()
            if "%s" in attr_name_start:
                res += attr_name_start % (chkr_name,)
            else:
                res += attr_name_start

            if attr is not None:
                if checker.optional:
                    res += "[" + attr
                    if checker._default is not UNDEFINED:
                        res += " (" + json2.dumps(checker._default) + ")"
                    res += "]"
                else:
                    res += attr
                res += ": "
            res += "<span>" + chkr_name + "</span>"
            if opt_end_inner:
                res += opt_end if checker.optional else mand_end
            res += attr_name_end
            if isinstance(checker, EnumChkr):
                res += (
                    enum_info_start
                    + json2.dumps(tuple(checker.options))
                    + enum_info_end
                )
            descr = getattr(checker, "_descr", None)
            if descr:
                res += arg_descr_start + checker._descr + arg_descr_end
            if checker.args:
                group = ""
                for value in checker.args:
                    if not isinstance(value, Checker):
                        continue
                    # endif
                    group += nl + checker_info(
                        None,
                        value,
                        indent=indent + 1,
                        indent_str=indent_str,
                    )
                group = group.strip()
                if group:
                    if "%s" in group_start:
                        res += group_start % (chkr_name,)
                    else:
                        res += group_start
                    res += group + group_end + nl + nl_indent

            if checker.kwargs:
                group = ""
                for key, value in checker.kwargs.items():
                    if not isinstance(value, Checker):
                        continue
                    # endif
                    group += nl + checker_info(
                        key,
                        value,
                        indent=indent + 1,
                        indent_str=indent_str,
                    )
                # endfor
                group = group.strip()
                if group:
                    if "%s" in group_start:
                        res += group_start % (chkr_name,)
                    else:
                        res += group_start
                    res += group + group_end + nl + nl_indent
            # endif

            if not opt_end_inner:
                res += opt_end if checker.optional else mand_end

            return res.rstrip()

        # enddef

        response = ""

        orig_mth = getattr(mth, "_wraps", mth)
        if orig_mth.md_help:
            doc = orig_mth.md_help
            if markdown and markdown_to_html:
                doc = markdown.markdown(
                    doc, extensions=md_extensions, output_format="html"
                )
                doc = '<div class="md-help">' + doc + "</div>"
        else:
            doc = orig_mth.__doc__
            if doc and nl != "\n":
                doc = doc.replace("\n", nl)
        if doc:

            response += (
                heading_start + trans.get("popis", "popis") + heading_end
            )
            response += nl + indent_str
            response += info_start + doc + info_end
            response += 2 * nl

        if devel:
            response += (
                heading_start + trans.get("vyskyt", "vyskyt") + heading_end
            )

            response += nl + indent_str
            response += (
                info_start
                + f"{inspect.getsourcefile(orig_mth)} "
                + f"(r.{inspect.getsourcelines(mth._wraps)[1]})"
                + info_end
            )
            response += 2 * nl

        response += (
            heading_start + trans.get("params", "params") + heading_end + nl
        )

        indent = 1
        i_group = ""
        for key, value in mth.deco_kwargs.items():
            if hasattr(value, "_doc_hidden") and value._doc_hidden is True:
                continue
            c_i = checker_info(key, value, indent=indent, indent_str=indent_str)
            i_group += c_i + nl
        # endfor
        i_group.strip()
        if i_group:
            if "%s" in group_start:
                response += group_start % ("root",)
            else:
                response += group_start
            response += i_group + group_end

        response += (
            nl
            + heading_start
            + trans.get("response", "response")
            + heading_end
            + nl
        )
        indent = 1
        i_group = ""
        for key, value in (
            ("status", IntChkr()),
            ("status_message", StrChkr()),
            (method, mth.response_format or AnyChkr(optional=True)),
        ):
            c_i = checker_info(key, value, indent=indent, indent_str=indent_str)
            i_group += c_i + nl
        # endfor
        i_group.strip()
        if i_group:
            if "%s" in group_start:
                response += group_start % ("resp",)
            else:
                response += group_start
            response += i_group + group_end
        response = response.strip()

        # vypis raises:

        if mth.raises:
            k = trans.get("raises", "raises")
            response += nl + heading_start + k + heading_end
            for e in mth.raises:
                response += nl + (
                    "<div>"
                    f"<b>{e.__name__}:</b> "
                    f"<i>HTTP:{e.http_status}</i> "
                    f"<i>Status:{e.status}</i> "
                    " - " + (e.__doc__ or "") + "</div>"
                )

        # Vypis other_checkers
        for k, v in mth.other_checkers or []:
            if hasattr(v, "_doc_hidden") and v._doc_hidden is True:
                continue
            response += nl + heading_start + k + ":" + heading_end + nl

            indent = 1
            i_group = ""
            c_i = checker_info(
                None, v or AnyChkr(), indent=indent, indent_str=indent_str
            )
            i_group += c_i + nl
            i_group.strip()
            if i_group:
                if "%s" in group_start:
                    response += group_start % ("resp",)
                else:
                    response += group_start
                response += i_group + group_end
            response = response.strip()
        # endfor

        if not response:
            raise MethodHelpNotFound(method)
        return response

    def openapi_v3(self):
        try:
            if yaml is None:
                return Str(
                    "Yaml not available! On server. Openapi doc not generated."
                )
            gen = DocGenOpenApi3()
            res = gen.process_server(
                {
                    "version": self.app.app_version,
                    "app_name": self.app.app_name,
                    "opeanapi_doc_description": self._get_doc_md_content(),
                    "methods": [
                        self.app._methods_info_method(i)
                        for i in sorted(self.app._methods.keys())
                        if not self._is_internal_method(i)
                    ],
                    "base_url": self.app.app_base_url,
                }
            )
            return Yaml(
                yaml.dump(res, allow_unicode=True, default_flow_style=False)
            )
        except Exception:
            log.exception("Error", exc_info=True)
            return Str("Error while generating openapi doc")
