from .rpcserver import Server
from .decorators import method
from .connectors import connect_postgre
from .checkers import IntChkr
from .exceptions import NotFound
from jsonrpc.decorators import cursor
from jsonrpc.config import load_config
from logging import getLogger
from functools import partial

log = getLogger('CRUD helper')


def dict_minus(dct, key):
    new = dct.copy()
    del new[key]
    return new


class CRUD(Server):
    def __init__(self,
                 name,
                 table=None,
                 columns=None,
                 key=None,
                 filters=None):
        Server.__init__(self)
        self.name = name

        self.config = load_config()
        self.columns = columns if columns else {}
        self.table = table if table else name
        self.key = key if key else "id"
        self.filters = filters if filters else self.columns.copy()
        self.filters['offset'] = IntChkr()
        self.filters['count'] = IntChkr()
        #  listovani
        setattr(
            self,
            name + "_list",
            method(
                self.filters
            )(
                lambda *args, **kwargs:
                    self._list(*args, **kwargs)
            )
        )
        getattr(
            self,
            name + "_list"
        ).__doc__ = "\n".join(
            [
                self.name + "_list",
                "Params:"
            ] + [
                "\t%s:\t%r" % fil
                for fil in self.filters.items()
            ]
        )

        #  vytvareni
        setattr(
            self,
            name + "_create",
            method(
                dict_minus(self.columns, self.key)
            )(
                lambda *args, **kwargs:
                    self._create(*args, **kwargs)
            )
        )
        getattr(
            self,
            name + "_create"
        ).__doc__ = "\n".join(
            [
                self.name + "_create",
                "Params:"
            ] + [
                "\t%s:\t%r" % fil
                for fil in self.columns.items()
                if fil[0] != self.key
            ]
        )

        #  update
        setattr(
            self,
            name + "_update",
            method(
                self.columns
            )(
                lambda *args, **kwargs:
                    self._update(*args, **kwargs)
            )
        )
        getattr(
            self,
            name + "_update"
        ).__doc__ = "\n".join(
            [
                self.name + "_update",
                "Params:"
            ] + [
                "\t%s:\t%r" % fil
                for fil in self.columns.items()
            ]
        )

        #  info
        setattr(
            self,
            name + "_info",
            method(
                {self.key: self.columns[self.key]}
            )(
                lambda *args, **kwargs:
                    self._info(*args, **kwargs)
            )
        )
        getattr(
            self,
            name + "_info"
        ).__doc__ = "\n".join(
            [
                self.name + "_info",
                "Params:",
                "\t%s:\t%r" % (self.key, self.columns[self.key])
            ]
        )

        #  delete
        setattr(
            self,
            name + "_delete",
            method(
                {self.key: self.columns[self.key]}
            )(
                lambda *args, **kwargs:
                    self._delete(*args, **kwargs)
            )
        )
        getattr(
            self,
            name + "_delete"
        ).__doc__ = "\n".join(
            [
                self.name + "_delete",
                "Params:",
                "\t%s:\t%r" % (self.key, self.columns[self.key])
            ]
        )

    @cursor(True, real_dict_cursor=True)
    def _list(self, c, offset=0, count=50, **kwargs):
        order_by = kwargs.pop('_order_by', [])
        nulls = kwargs.pop('_nulls', None)
        cols = list(kwargs.keys())  # pro zachovani poradi
        qry = ["select"]
        qry += [",".join(self.columns.keys())]
        qry += ["from", self.table]
        qry += ["where"] if cols else []
        qry += [" and ".join(["%s = %%s" % (col, ) for col in cols])]
        qry += ["order by"] if order_by else []
        qry += [" , ".join([
            "%s %s" % (col[0], col[1])
            for col in order_by
            if col[0] in self.columns and col[1] in ("asc", "desc")
        ])]
        qry += ["nulls %s" % (nulls,)] if \
            nulls and order_by and nulls in ("first", "last") else []
        qry += ["limit %s offset %s"]
        c.execute(
            " ".join(qry),
            [kwargs[col] for col in cols] + [count, offset]
        )
        for row in c:
            yield row

    @cursor(True, real_dict_cursor=True)
    def _create(self, c, **kwargs):
        cols = list(kwargs.keys())  # pro zachovani poradi
        qry = ["insert into", self.table, "("]
        qry += [", ".join(cols)]
        qry += [") values ("]
        qry += [", ".join(["%s"] * len(cols))]
        qry += [") returning", self.key]
        c.execute(
            " ".join(qry),
            [kwargs[k] for k in cols]
        )
        res = c.fetchone()
        return res

    @cursor(True, real_dict_cursor=True)
    def _update(self, c, **kwargs):
        cols = list(kwargs.keys())  # pro zachovani poradi
        dml = ["update", self.table, "set"]
        fields = ["%s = %%s" % col for col in cols if col[0] != self.key]
        where = ["where %s=%%s" % (self.key, )]

        qry = self.join(" ", dml)
        qry += " " + self.join(", ", fields)
        qry += self.join(" ", where)

        c.execute(
            qry,
            [kwargs[k] for k in cols] + [kwargs[self.key]]
        )
        return c.rowcount

    @cursor(True, real_dict_cursor=True)
    def _info(self, c, **kwargs):
        qry = ["select"]
        qry += [",".join(self.columns.keys())]
        qry += ["from", self.table]
        qry += ["where", self.key, "= %s ", ]
        c.execute(
            " ".join(qry),
            (kwargs[self.key],)
        )
        res = c.fetchone()
        if not res:
            raise NotFound()

        return res

    @cursor(True, real_dict_cursor=True)
    def _delete(self, c, **kwargs):
        qry = "delete from %s where %s=%%s" % (self.table, self.key, )
        qry_args = (kwargs[self.key], )
        c.execute(qry, qry_args)
        return True

    def join(self, connector, sequence):
        return connector.join(map(str, sequence))

    def postfork(self):
        self.postgre = connect_postgre(self.config)


class MultiCRUD(Server):
    def __init__(self, mapping, *args, **kwargs):
        Server.__init__(self, *args, **kwargs)
        """
mapping je externi_jmeno => {"table", "columns", "key", "filters"}
        """
        self.mapping = mapping

        self.config = load_config()

        def get_key(name):
            return self.mapping[name].get("key") or "id"

        def get_cols(name):
            return self.mapping[name].get("columns") or {}

        def get_fltrs(name):
            filters = self.mapping[name].get(
                "filters"
            ) or get_cols(name).copy()
            filters['offset'] = IntChkr()
            filters['count'] = IntChkr()
            return filters

        for name, mapp in self.mapping.items():
            #  listovani
            setattr(
                self,
                name + "_list",
                partial(
                    method(
                        get_fltrs(name)
                    )(lambda *args, **kwargs: self._list(*args, **kwargs)),
                    mapp=mapp
                )
            )
            m = getattr(
                self,
                name + "_list"
            )
            m.__doc__ = "\n".join(
                [
                    name + "_list",
                    "Params:"
                ] + [
                    "\t%s:\t%r" % fil
                    for fil in get_fltrs(name).items()
                ]
            )
            m.exposed = True

            #  vytvareni
            setattr(
                self,
                name + "_create",
                partial(
                    method(
                        dict_minus(get_cols(name), get_key(name))
                    )(lambda *args, **kwargs: self._create(*args, **kwargs)),
                    mapp=mapp
                )
            )
            m = getattr(
                self,
                name + "_create"
            )
            m.__doc__ = "\n".join(
                [
                    name + "_create",
                    "Params:"
                ] + [
                    "\t%s:\t%r" % fil
                    for fil in get_cols(name).items()
                    if fil[0] != get_key(name)
                ]
            )
            m.exposed = True

            #  update
            setattr(
                self,
                name + "_update",
                partial(
                    method(
                        get_cols(name)
                    )(lambda *args, **kwargs: self._update(*args, **kwargs)),
                    mapp=mapp
                )
            )
            m = getattr(
                self,
                name + "_update"
            )
            m.__doc__ = "\n".join(
                [
                    name + "_update",
                    "Params:"
                ] + [
                    "\t%s:\t%r" % fil
                    for fil in get_cols(name).items()
                ]
            )
            m.exposed = True

            #  info
            setattr(
                self,
                name + "_info",
                partial(
                    method(
                        {get_key(name): get_cols(name)[get_key(name)]}
                    )(lambda *args, **kwargs: self._info(*args, **kwargs)),
                    mapp=mapp
                )
            )
            m = getattr(
                self,
                name + "_info"
            )
            m.__doc__ = "\n".join(
                [
                    name + "_info",
                    "Params:",
                    "\t%s:\t%r" % (
                        get_key(name), get_cols(name)[get_key(name)])
                ]
            )
            m.exposed = True

            #  delete
            setattr(
                self,
                name + "_delete",
                partial(
                    method(
                        {get_key(name): get_cols(name)[get_key(name)]}
                    )(lambda *args, **kwargs: self._delete(*args, **kwargs)),
                    mapp=mapp
                )
            )
            m = getattr(
                self,
                name + "_delete"
            )
            m.__doc__ = "\n".join(
                [
                    name + "_delete",
                    "Params:",
                    "\t%s:\t%r" % (
                        get_key(name), get_cols(name)[get_key(name)])
                ]
            )
            m.exposed = True

    @cursor(True, real_dict_cursor=True)
    def _list(self, c, mapp, offset=0, count=50, **kwargs):
        order_by = kwargs.pop('_order_by', [])
        nulls = kwargs.pop('_nulls', None)
        cols = list(kwargs.keys())  # pro zachovani poradi
        qry = ["select"]
        columns = mapp.get("columns") or {}
        table = mapp["table"]
        qry += [",".join(columns.keys())]
        qry += ["from", table]
        qry += ["where"] if cols else []
        qry += [" and ".join(["%s = %%s" % (col, ) for col in cols])]
        qry += ["order by"] if order_by else []
        qry += [" , ".join([
            "%s %s" % (col[0], col[1])
            for col in order_by
            if col[0] in columns and col[1] in ("asc", "desc")
        ])]
        qry += ["nulls %s" % (nulls,)] if \
            nulls and order_by and nulls in ("first", "last") else []
        qry += ["limit %s offset %s"]
        c.execute(
            " ".join(qry),
            [kwargs[col] for col in cols] + [count, offset]
        )
        for row in c:
            yield row

    @cursor(True, real_dict_cursor=True)
    def _create(self, c, mapp, **kwargs):
        key = mapp.get("key") or "id"
        table = mapp["table"]
        cols = list(kwargs.keys())  # pro zachovani poradi
        qry = ["insert into", table, "("]
        qry += [", ".join(cols)]
        qry += [") values ("]
        qry += [", ".join(["%s"] * len(cols))]
        qry += [") returning", key]
        c.execute(
            " ".join(qry),
            [kwargs[k] for k in cols]
        )
        res = c.fetchone()
        return res

    @cursor(True, real_dict_cursor=True)
    def _update(self, c, mapp, **kwargs):
        cols = list(kwargs.keys())  # pro zachovani poradi
        key = mapp.get("key") or "id"
        table = mapp["table"]
        dml = ["update", table, "set"]
        fields = ["%s = %%s" % col for col in cols if col[0] != key]
        where = ["where", key, "=", kwargs[key]]

        qry = self.join(" ", dml)
        qry += " " + self.join(", ", fields)
        qry += self.join(" ", where)

        c.execute(
            qry,
            [kwargs[k] for k in cols]
        )
        return c.rowcount

    @cursor(True, real_dict_cursor=True)
    def _info(self, c, mapp, **kwargs):
        columns = mapp.get("columns") or {}
        key = mapp.get("key") or "id"
        table = mapp["table"]
        qry = ["select"]
        qry += [",".join(columns.keys())]
        qry += ["from", table]
        qry += ["where", key, "= %s ", ]
        c.execute(
            " ".join(qry),
            (kwargs[key],)
        )
        res = c.fetchone()
        if not res:
            raise NotFound()

        return res

    @cursor(True, real_dict_cursor=True)
    def _delete(self, c, mapp, **kwargs):
        key = mapp.get("key") or "id"
        table = mapp["table"]
        qry = ["delete from", table, "where"]
        qry += [key, "=", str(kwargs[key])]
        c.execute(
            " ".join(qry),
            (kwargs[key],)
        )
        return True

    def join(self, connector, sequence):
        return connector.join(map(str, sequence))

    def postfork(self):
        self.postgre = connect_postgre(self.config)
