from logging import getLogger
import socket
from datetime import timedelta
from time import time, sleep

from jsonrpc.connectors import connect_postgre
from jsonrpc.connectors import connect_redis
from jsonrpc.decorators import cursor

log = getLogger("jsonrpc.master_checker")


class MasterChecker(object):
    def __init__(self, config, section="master_checker"):
        self.config = config
        self.section = section
        self.keepalive_period = config.getint(section, "keepalive", 10)
        self.master_expire = timedelta(seconds=config.getint(
            section, "master_expire", 300))
        self.ident = "%s-%s" % (
            socket.getfqdn(),
            config.get(section, "ident")
        )
        self.master = False
        self.running = False
        self.should_run = False
    # enddef

    def stop(self):
        self.should_run = False

    def _update_keepalive(self, *args, **kwargs):
        raise NotImplementedError()

    def keepalive(self, *args, **kwargs):
        try:
            self._update_keepalive(*args, **kwargs)
        except BaseException as e:
            log.error("Chyba keepalive:%s-%s", e.__class__.__name__, e)
            self.master = False
            return
        # enddef
        self.check_master(*args, **kwargs)
    # enddef

    def _remove_all_masters(self, *args, **kwargs):
        raise NotImplementedError()
    # enddef

    def _set_master(self, *args, **kwargs):
        raise NotImplementedError()
    # enddef

    def check_master(self, *args, **kwargs):
        masters = []
        m_len = 0
        try:
            while m_len != 1:
                masters = self._get_masters(*args, **kwargs)
                m_len = len(masters)
                if m_len > 1:
                    self._remove_all_masters(*args, **kwargs)
                elif m_len == 0:
                    self._set_master(*args, **kwargs)
                # endif
            # endwhile
        except BaseException as e:
            log.error("Chyba check_master:%s-%s", e.__class__.__name__, e)
            masters = []
        # endfor
        self.master = masters and (masters[0] == self.ident)
        return self.master

    def _get_masters(self, *args, **kwargs):
        raise NotImplementedError()
    # enddef

    def sign_in(self, *args, **kwargs):
        self._sign_in(*args, **kwargs)
    # enddef

    def _sign_in(self, *args, **kwargs):
        pass
    # enddef

    def run(self, method, *args, **kwargs):
        self.should_run = True
        last_keepalive = 0
        self.sign_in()
        while self.should_run:
            t = time()
            if t - self.keepalive_period > last_keepalive:
                last_keepalive = t
                self.keepalive()
            if self.master:
                method(*args, **kwargs)
            else:
                sleep(self.keepalive_period)
            # endif
        # endwhile
        self.running = False
    # enddef
# enddef


class MasterCheckerDuality(MasterChecker):

    def __init__(self, *args, **kwargs):
        super(MasterCheckerDuality, self).__init__(*args, **kwargs)
        self.is_duality_master = self.config.getboolean(
            "uwsgi", "duality_master", False
        )
    # enddef

    def _update_keepalive(self, *args, **kwargs):
        pass

    def _get_masters(self, *args, **kwargs):
        if self.is_duality_master:
            return [self.ident, ]
        return ["default_master", ]
    # enddef

    def _remove_all_masters(self, *args, **kwargs):
        return []
    # enddef

    def _set_master(self, *args, **kwargs):
        pass
    # enddef

# endclass


class MasterCheckerDB(MasterChecker):

    def __init__(self, *args, **kwargs):
        super(MasterCheckerDB, self).__init__(*args, **kwargs)
        self.lock_table = self.config.get(self.section, "lock_table", "master")
        self.group = self.config.get(self.section, "group", None)
    # enddef

    def _commit(self):
        self.postgre.commit()

    def _get_masters(self, cursor):
        qargs = [self.master_expire, ]
        q = "select ident from " + self.lock_table + """
            where master is TRUE and last_update > (NOW() - %s )"""
        if self.group:
            q += """ and process_group=%s"""
            qargs.append(self.group)
        # endif

        cursor.execute(q, qargs)
        return [i[0] for i in cursor]
    # enddef

    def _update_keepalive(self, cursor, *args, **kwargs):
        q = "update " + self.lock_table + """
            set last_update = NOW() where ident = %s """
        qargs = [self.ident, ]
        if self.group:
            q += """ and process_group=%s"""
            qargs.append(self.group)
        # endif
        q += "returning *"
        cursor.execute(q, qargs)
        row = cursor.fetchone()
        if not row:
            q = "insert into " + self.lock_table + """ (ident, last_update%s)
                VALUES (%%s, NOW()%s)"""
            if self.group:
                q %= (", process_group", ",%s")
            # endif
            q += "returning *"
            cursor.execute(q, qargs)
            row = cursor.fetchone()
        # endif
        self._commit()

    def _remove_all_masters(self, cursor, *args, **kwargs):
        q = "update " + self.lock_table + " set master = FALSE"
        qargs = []
        if self.group:
            q += " where process_group = %s"
            qargs.append(self.group)
        # endif
        cursor.execute(q, qargs)
        self._commit()
    # enddef

    def _set_master(self, cursor, *args, **kwargs):
        q = "update " + self.lock_table + """
            set master = TRUE where ident=%s"""
        qargs = [self.ident, ]
        if self.group:
            q += """ and process_group=%s"""
            qargs.append(self.group)
        # endif
        cursor.execute(q, qargs)
        self._commit()
    # enddef

    @cursor
    def keepalive(self, cursor, *args, **kwargs):
        super(MasterCheckerDB, self).keepalive(
            *args, cursor=cursor, **kwargs
        )
    # enddef

    @cursor
    def check_master(self, cursor, *args, **kwargs):
        super(MasterCheckerDB, self).check_master(
            *args, cursor=cursor, **kwargs
        )
    # enddef

    @cursor
    def _sign_in(self, cursor, *args, **kwargs):
        q = "update " + self.lock_table + """
            set master = FALSE where ident=%s"""
        qargs = [self.ident, ]
        if self.group:
            q += """ and process_group=%s"""
            qargs.append(self.group)
        # endif
        cursor.execute(q, qargs)
        self._commit()
    # enddef


class MasterCheckerPsql(MasterCheckerDB):
    def __init__(self, *args, **kwargs):
        psql_section = kwargs.pop("psql_section", None)
        super(MasterCheckerPsql, self).__init__(*args, **kwargs)
        self.postgre = connect_postgre(
            config=self.config, section=psql_section
        )
    # enddef

    def _commit(self):
        self.postgre.commit()
    # enddef

# endclass


class MasterCheckerRedis(MasterChecker):
    def __init__(self, *args, **kwargs):
        redis_section = kwargs.pop("redis_section", None)
        super(MasterCheckerRedis, self).__init__(*args, **kwargs)
        self.redis = connect_redis(
            config=self.config, section=redis_section
        )
        self.lock_key = self.config.get(self.section, "lock_key", "master")
    # enddef

    def _get_masters(self):
        v = self.redis.get(self.lock_key)
        if v:
            return [v]
        return []
    # enddef

    def _update_keepalive(self, *args, **kwargs):
        m = self._get_masters()
        if m and m[0] == self.ident:
            success = self.redis.expire(
                self.lock_key, int(self.master_expire.total_seconds())
            )
            assert success, "Nepodařilo se prodloužit expiraci klíče"
        # endif
    # enddef

    def _remove_all_masters(self, *args, **kwargs):
        success = self.redis.delete(self.lock_key)
        assert success
    # enddef

    def _set_master(self, *args, **kwargs):
        success = self.redis.setnx(self.lock_key, self.ident)
        assert success, "Nejsem master!"
        if success:
            success = self.redis.expire(
                self.lock_key, int(self.master_expire.total_seconds())
            )
            assert success, "Nepodařilo se prodloužit expiraci klíče"
        # endif
    # enddef


# endclass

# eof
