import paramiko
from socket import timeout as SocketTimeout
import os
import logging
import time
import re


# this is quote from the shlex module, added in py3.3
_find_unsafe = re.compile(r'[^\w@%+=:,./~-]').search


def _sh_quote(s):
    """Return a shell-escaped version of the string *s*."""
    if not s:
        return ""
    if _find_unsafe(s) is None:
        return s

    # use single quotes, and put single quotes into double quotes
    # the string $'b is then quoted as '$'"'"'b'
    return "'" + s.replace("'", "'\"'\"'") + "'"
# enddef


class SCP(object):
    """
    SCP class
    """
    def __init__(
        self, host=None, port=22, config=None, config_section=None, user=None,
        password=None, progress=None, timeout=5.0, bufferSize=16384
    ):
        super(SCP, self).__init__()

        # load configuration
        if host:
            self.host = host
            self.port = port
            self.user = user
            self.password = password
            self.socketTimeout = timeout
            self.bufferSize = bufferSize

        elif config and config_section:
            self.host = config.get(config_section, "host")
            self.port = config.getint(config_section, "port", 22)
            self.user = config.get(config_section, "user", None)
            self.password = config.get(config_section, "password", None)
            self.socketTimeout = config.getfloat(
                config_section, "timeout", 5.0
            )
            self.bufferSize = config.get(config_section, "bufferSize", 16384)

        else:
            raise SCPException("SCP not configured!")
        # endif

        self._progress = progress

        self.__createTransport()

        (stdin, stdout, stderr) = self.client.exec_command("uname -a")
        stdin.close()
        response = stdout.read()
        err = stderr.read()
        logging.debug("Connected to:%s", response)
        if err:
            logging.warning("SSH stderr:%s", err)
        # endif
        self.channel = None
    # enddef

    def close(self):
        if self.channel:
            self.channel.close()
            self.channel = None
        # endif
        if self.transport:
            self.transport.close()
            self.transport = None
        # endif
        if self.client:
            self.client.close()
            self.client = None
        # endif
    # enddef

    def __del__(self):
        """Pokus o zachranu pokud nejsou uzavrena spojeni....
        """
        self.close()
    # endef

    def sendFile(self, f, remote_path='.', fsize=None, fname=None):
        self.iterator = 0
        """
        Transfer file to remote host.

        @param f: A single path, or object with functions:
                    'read', 'tell' and 'close'
        @type files: string OR object
        @param remote_path: path in which to receive the files on the remote
            host. defaults to '.'
        @type remote_path: str
        @param fsize: Used when f is object. Represents filesize in B
        @type fsize: int
        @param fname: Used when f is object.
                        Filename if not provided by 'name' function on f.
        @type fname: str
        """
        # py3: str or bytes?
        if isinstance(f, str):
            filename = f
            fds = None
            fsize = None
        elif hasattr(f, "read") and hasattr(f, "close") and hasattr(f, "tell"):
            filename = None
            if not fname:
                if hasattr(f, "filename"):
                    fname = f.filename
                    print(fname, type(fname))
                    # if type(fname) != str:
                    #     raise SSHException(
                    #         "f.name must be str not %s"%(type(fname),)
                    #     )

                    # f.save('/tmp/' + fname)
                    fsize = os.stat('/tmp/' + fname).st_size
                    f = open('/tmp/' + fname, "rb")
                    print("fsize :", fsize)
                else:
                    raise SSHException(
                        "Provide fname if f hasn't atrribute name"
                    )
            fds = f
            # fsize = fsize
        else:
            raise SCPException(
                "Can't transfer object of type '%s'" % (type(f),)
            )
        # endif

        self.channel = self.transport.open_session()
        self.channel.settimeout(self.socketTimeout)
        scp_command = 'scp -t %s'
        self.channel.exec_command(scp_command % _sh_quote(remote_path))

        self.__recvConfirm()

        self.__sendFile(
            filename=filename, fds=fds, fsize=fsize, ftargetname=fname
        )

        if self.channel:
            self.channel.close()
        # endif
    # enddef

    def __createTransport(self):
        client = paramiko.SSHClient()
        client.load_system_host_keys()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        logging.debug(
            "Connecting to %s@%s:%d %s password",
            self.user, self.host, self.port,
            "with" if self.password else "without"
        )

        res = client.connect(self.host, self.port, self.user, self.password)
        logging.debug("SCP response:%s", res)
        self.client = client
        self.transport = client.get_transport()
    # endef

    def __readStats(self, name):
        """return just the file stats needed for scp"""
        try:
            stats = os.stat(name)
            mode = oct(stats.st_mode)[-4:]
            size = stats.st_size
            atime = int(stats.st_atime)
            mtime = int(stats.st_mtime)
        except (OSError, TypeError):
            mode = "0664"
            size = None
            atime = mtime = int(time.time())
        # endtry

        return (mode, size, mtime, atime)
    # endef

    def __recvConfirm(self):
        self.iterator += 1
        print("recvConfirm")
        # read scp response
        msg = ''
        try:
            msg = self.channel.recv(512)
        except SocketTimeout:
            raise SCPException('Timout waiting for scp response')
        # slice off the first byte,
        # so this compare will work in python2 and python3
        print(self.iterator)
        if msg and msg[0:1] == b'\x00':
            return
        elif msg and msg[0:1] == b'\x01':
            raise SCPException(msg[1:])
        elif self.channel.recv_stderr_ready():
            msg = self.channel.recv_stderr(512)
            raise SCPException(msg)
        elif not msg:
            raise SCPException('No response from server')
        else:
            raise SCPException('Invalid response from server', msg)
        # endif
    # enddef

    def __sendTime(self, mtime, atime):
        self.channel.sendall('T%d 0 %d 0\n' % (mtime, atime))
        self.__recvConfirm()
    # endef

    def __sendFile(
        self, filename=None, fds=None, fsize=None, ftargetname=None
    ):
        if filename and not fds:
            fds = open(filename, "rb")
        # endif
        localname = filename
        if fds and fsize is None and hasattr(fds, "name"):
            localname = fds.name
        # endif
        if ftargetname is not None:
            basename = os.path.basename(ftargetname)
        else:
            basename = os.path.basename(filename)
        # endif

        (mode, size, mtime, atime) = self.__readStats(localname)
        if fds and size is None:
            if fsize is None:
                raise SCPException("Providing fds needs setting also fsize")
            # endif
            size = fsize
        # endif

        self.__sendTime(mtime, atime)

        # The protocol can't handle \n in the filename.
        # Quote them as the control sequence \^J for now,
        # which is how openssh handles it.
        print(
            'just sending file ',
            ("C%s %d %s\n" % (mode, size, basename.replace('\n', '\\^J')))
        )
        self.channel.sendall("C%s %d %s\n" %
                             (mode, size, basename.replace('\n', '\\^J')))
        print('sending end')
        self.__recvConfirm()
        file_pos = 0
        if self._progress:
            if size == 0:
                # avoid divide-by-zero
                self._progress(basename, 1, 1)
            else:
                self._progress(basename, size, 0)
            # endif
        # endif
        buff_size = self.bufferSize
        chan = self.channel
        while file_pos < size:
            chan.sendall(fds.read(buff_size))
            file_pos = fds.tell()
            if self._progress:
                self._progress(basename, size, file_pos)
            # endif
        # endwhile
        chan.sendall('\x00')
        fds.close()
        self.__recvConfirm()
    # enddef
# endclass


class SCPException(Exception):
    """SCP exception class"""
    pass
# endclass


if __name__ == "__main__":
    # log to stderr
    logger = logging.getLogger()
    logger.addHandler(logging.StreamHandler())
    # log everything
    logger.setLevel(0)

    # connect
    s = SCP(
        host="lukoko.ats",
        user="root",
        password="",
    )
    # def sendFile(self, f, remote_path='.', fsize = None, fname = None):
    # poslu lokalni soubor
    s.sendFile("/etc/hosts", remote_path="/tmp/")

    s.sendFile("/etc/hosts", remote_path="/tmp/", fname="hostscopy")

    # poslu uz otevreny file descriptor na existujici soubor
    fds = open("/etc/resolv.conf", "rb")
    s.sendFile(fds, remote_path="/tmp/", fname="darkyresolv")

    # poslu data z pameti
    from io import StringIO
    data = StringIO("""sfjzsdgfdhgdh
dfg
j
dfgh
j
ghf
j
fgh

k
ghj
kl
ghjklhjkl

jk
lhjkkkj""")

    s.sendFile(data, remote_path="/tmp/", fname="balast", fsize=data.len)
    s.close()
